import numpy as np

from pyaerocom._lowlevel_helpers import invalid_input_err_str
from pyaerocom.colocation.colocation_utils import _colocate_site_data_helper
from pyaerocom.geodesy import find_coord_indices_within_distance
from pyaerocom.helpers import sort_ts_types
from pyaerocom.obs_io import ObsVarCombi
from pyaerocom.stationdata import StationData


def _check_input_data_ids_and_vars(data_ids_and_vars):
    if not isinstance(data_ids_and_vars, list | tuple):
        raise ValueError("Input data_ids_and_vars must be tuple or list")
    elif len(data_ids_and_vars) != 2:
        raise NotImplementedError("Currently, only (and exactly) 2 datasets can be combined...")
    for item in data_ids_and_vars:
        if not isinstance(item, list | tuple):
            raise ValueError("Each entry in data_ids_and_vars must be tuple or list")
        elif len(item) != 3:
            raise ValueError("Each entry in data_ids_and_vars needs to contain exactly 3 items.")
        if not isinstance(item[1], str) or not isinstance(item[2], str):
            raise ValueError("2nd and 3rd entries (data_id, var_name) in item need to be str")


def _map_same_stations(stats_short, stats_long, match_stats_how, match_stats_tol_km):
    long_coords = list(zip(stats_long["latitude"], stats_long["longitude"]))

    # index matches and corresponding station name matches
    _index_short = []
    _index_long = []
    _statnames_short = []
    _statnames_long = []

    long_sitenames = np.asarray(stats_long["station_name"])

    for i, stat in enumerate(stats_short["stats"]):
        statname = stat.station_name
        lat0, lon0 = stats_short["latitude"][i], stats_short["longitude"][i]

        if match_stats_how == "station_name":
            # np.where returns tuple, first index contains array with index
            # matches
            index_matches = np.where(long_sitenames == statname)[0]
        else:
            index_matches = find_coord_indices_within_distance(
                latref=lat0, lonref=lon0, latlons=long_coords, radius=match_stats_tol_km
            )

        # init which default index to use
        use_index = 0
        if len(index_matches) == 0:
            continue
        elif len(index_matches) > 1:
            if match_stats_how == "station_name":
                raise Exception(
                    "Unexpected error: each station_name should "
                    "only occur once... (perhaps due to unforeseen "
                    "API change sometime in the future)"
                )
            else:
                # more than one site was found in the surroundings of the
                # current coordinate. Check and prefer same site name if
                # possible, else, use closest
                for j, idx_match in enumerate(index_matches):
                    if statname == stats_long["station_name"][idx_match]:
                        use_index = j
                        break

        idx_long = index_matches[use_index]

        # make sure to colocate each site only once
        if idx_long in _index_long:
            statname_long = stats_long["station_name"][idx_long]
            if statname == statname_long:
                # rare case: the index match in long has already been assigned
                # to another site in short which does not occur in long
                # (e.g. AAOT site and Venise site in AERONET). In this case
                # we want to use the one that matches the site name, so we
                # have to remove the already registered index from the record
                rm_idx = _statnames_long.index(statname_long)
                _index_short.pop(rm_idx)
                _index_long.pop(rm_idx)
                _statnames_long.pop(rm_idx)
                _statnames_short.pop(rm_idx)
            else:
                continue

        _index_short.append(i)
        _index_long.append(idx_long)
        _statnames_short.append(statname)
        _statnames_long.append(stats_long["station_name"][idx_long])

    return (_index_short, _index_long, _statnames_short, _statnames_long)


def _combine_2_sites(
    stat,
    var,
    stat_other,
    var_other,
    merge_how,
    merge_eval_fun,
    match_stats_tol_km,
    var_name_out,
    data_id_out,
    var_unit_out,
    resample_how,
    min_num_obs,
    prefer,
    merge_info_vars,
    add_meta_keys,
):
    """Combine two StationData objects for a given merge strategy

    Private for now...  details should follow. Until then see
    :func:`combine_vardata_ungridded` for details on input args

    Returns
    -------
    StationData
        merged StationData instance
    """
    # unit of first variable
    var_unit_in = stat.get_unit(var)

    # check if output unit is defined explicitly and if not, use unit of
    # variable 1
    if var_unit_out is None:
        var_unit_out = var_unit_in
    # make sure both input data objects are in the correct unit (which is
    # var_unit_out)
    elif not var_unit_in == var_unit_out:
        stat.convert_unit(var, var_unit_out)

    if not stat_other.get_unit(var_other) == var_unit_out:
        stat_other.convert_unit(var_other, var_unit_out)

    new = StationData()
    # add default metadata to new data object
    meta_first = stat.get_meta(
        force_single_value=False, quality_check=False, add_meta_keys=add_meta_keys
    )
    new.update(meta_first)

    new.merge_meta_same_station(
        other=stat_other,
        check_coords=False,  # has already been done
        inplace=True,
        raise_on_error=True,
        add_meta_keys=add_meta_keys,
    )

    tstype = stat.get_var_ts_type(var)
    tstype_other = stat_other.get_var_ts_type(var_other)

    to_ts_type = sort_ts_types([tstype, tstype_other])[-1]

    df = _colocate_site_data_helper(
        stat,
        stat_other,
        var,
        var_other,
        to_ts_type,
        resample_how=resample_how,
        min_num_obs=min_num_obs,
        use_climatology_ref=False,
    )

    # remove timestamps where both observations are NaN
    df.dropna(axis=0, how="all", inplace=True)

    # NOTE: the dataframe returned by _colocate_site_data_helper has ref as first
    # column and the first input data as 2nd!
    obsvar_id = str(ObsVarCombi(stat.data_id, var))
    obsvar_id_other = str(ObsVarCombi(stat_other.data_id, var_other))

    stat_order = [stat_other, stat]
    col_order = [obsvar_id_other, obsvar_id]
    col_vars = [var_other, var]
    col_names = list(df.columns.values)

    # In case input variables are different, keep both of them in the
    # output colocated StationData, in addition to potentially computed
    # additional variables below. This is equivalent with using
    # merge_how='combine' and var1 != var2
    if var != var_other:
        for j, colname in enumerate(col_names):
            _var = col_vars[j]
            _stat = stat_order[j]
            ts = df[colname]
            new[_var] = ts
            vi = _stat["var_info"][_var]
            vi["ts_type"] = to_ts_type
            new["var_info"][_var] = vi

    add_ts = None
    # Merge timeseries if variables are the same and are supposed to be
    # combined
    if merge_how == "combine" and var == var_other:
        prefer_col = col_names[col_order.index(prefer)]
        dont_prefer = col_names[int(not (col_names.index(prefer_col)))]
        add_ts = df[prefer_col].combine_first(df[dont_prefer])

        if var_name_out is None:
            var_name_out = var

    elif merge_how == "mean":
        if var != var_other:
            raise NotImplementedError(
                "Averaging of site data is only supported if input variables are the same..."
            )
        # if it made it until here, then both sites have same variables and
        # units
        if var_name_out is None:
            var_name_out = var

        add_ts = df.mean(axis=1)

    elif merge_how == "eval":
        func = merge_eval_fun.replace(col_order[0], col_names[0])
        func = func.replace(col_order[1], col_names[1])
        if "=" in merge_eval_fun:
            # make sure variable name is not in merge_eval_fun anymore, otherwise
            # the eval method will return a DataFrame instead of a Series
            func = func.split("=")[-1].strip()
        add_ts = df.eval(func)

        if var_name_out is None:
            var_name_out = merge_eval_fun
            var_name_out = var_name_out.replace(f"{stat.data_id};", "")
            var_name_out = var_name_out.replace(f"{stat_other.data_id};", "")

    if add_ts is not None:
        var_info = {"ts_type": to_ts_type, "units": var_unit_out}

        var_info.update(merge_info_vars)

        new["var_info"][var_name_out] = var_info
        new[var_name_out] = add_ts

    if isinstance(data_id_out, str):
        new["data_id"] = data_id_out

    return new


def combine_vardata_ungridded(
    data_ids_and_vars,
    match_stats_how="closest",
    match_stats_tol_km=1,
    merge_how="combine",
    merge_eval_fun=None,
    var_name_out=None,
    data_id_out=None,
    var_unit_out=None,
    resample_how=None,
    min_num_obs=None,
    add_meta_keys=None,
):
    """
    Combine and colocate different variables from UngriddedData

    This method allows to combine different variable timeseries from different
    ungridded observation records in multiple ways. The source data may be all
    included in a single instance of `UngriddedData` or in multiple, for
    details see first input parameter :param:`data_ids_and_vars`. Merging can
    be done in flexible ways, e.g. by combining measurements of the same
    variable from 2 different datasets or by computing new variables based
    on 2 measured variables (e.g. concox=concno2+conco3). Doing this requires
    colocation of site locations and timestamps of both input observation
    records, which is done in this method.

    It comprises 2 major steps:

        1. Compute list of :class:`StationData` objects for both input \
            data combinations (data_id1 & var1; data_id2 & var2) and based \
            on these, find the coincident locations. Finding coincident \
            sites can either be done based on site location name or based on
            their lat/lon locations. The method to use can be specified via
            input arg :param:`match_stats_how`.
        2. For all coincident locations, a new instance of :class:`StationData` \
            is computed that has merged the 2 timeseries in the way
            that can be specified through input args :param:`merge_how` and
            :param:`merge_eval_fun`. If the 2 original timeseries from both
            sites come in different temporal resolutions, they will be
            resampled to the lower of both resolutions. Resampling constraints
            that are supposed to be applied in that case can be provided via
            the respective input args for temporal resampling. Default is
            pyaerocom default, which corresponds to ~25% coverage constraint
            (as of 22.10.2020) for major resolution steps, such as
            daily->monthly.

    Note
    ----
    Currently, only 2 variables can be combined to a new one (e.g.
    concox=conco3+concno2).

    Note
    ----
    Be aware of unit conversion issues that may arise if your input data is
    not in AeroCom default units. For details see below.

    Parameters
    ----------
    data_ids_and_vars : list
        list of 3 element tuples, each containing, in the following order
        1. instance of :class:`UngriddedData`; 2. dataset ID (remember that
        UngriddedData can contain more than one dataset); and 3. variable name.
        Note that currently only 2 of such tuples can be combined.
    match_stats_how : str, optional
        String specifying how site locations are supposed to be matched.
        The default is 'closest'. Supported are 'closest' and 'station_name'.
    match_stats_tol_km : float, optional
        radius tolerance in km for matching site locations when using 'closest'
        for site location matching. The default is 1.
    merge_how : str, optional
        String specifying how to merge variable data at site locations.
        The default is 'combine'. If both input variables are the same and
        `combine` is used, then the first input variable will be preferred over
        the other. Supported are 'combine', 'mean' and 'eval', for the latter,
        `merge_eval_fun` needs to be specified explicitly.
    merge_eval_fun : str, optional
        String specifying how `var1` and `var2` data should be evaluated (only
        relevant if `merge_how='eval'` is used) . The default is None. E.g. if
        one wants to retrieve the column aerosol fine mode fraction at 550nm
        (fmf550aer) through AERONET, this could be done through the SDA product
        by prodiding data_id1 and var1 are 'AeronetSDA' and 'od550aer' and
        second input data_id2 and var2 are 'AeronetSDA' and 'od550lt1aer' and
        merge_eval_fun could then be
        'fmf550aer=(AeronetSDA;od550lt1aer/AeronetSDA;od550aer)*100'. Note that
        the input variables will be converted to their AeroCom default units,
        so the specification of `merge_eval_fun` should take that into account
        in case the originally read obsdata is not in default units.
    var_name_out : str, optional
        Name of output variable. Default is None, in which case it is attempted
        to be inferred.
    data_id_out : str, optional
        `data_id` set in output `StationData` objects. Default is None, in
        which case it is inferred from input data_ids (e.g. in above example
        of merge_eval_fun, the output data_id would be 'AeronetSDA' since both
        input IDs are the same.
    var_unit_out : str
        unit of output variable.
    resample_how : str, optional
        String specifying how temporal resampling should be done. The default
        is 'mean'.
    min_num_obs : int or dict, optional
        Minimum number of observations for temporal resampling.
        The default is None in which case pyaerocom default is used, which
        is available via pyaerocom.const.OBS_MIN_NUM_RESAMPLE.
    add_meta_keys : list, optional
        additional metadata keys to be added to output `StationData` objects
        from input data. If None, then only the pyaerocom default keys are
        added (see `StationData.STANDARD_META_KEYS`).

    Raises
    ------
    ValueError
        If input for `merge_how` or `match_stats_how` is invalid.
    NotImplementedError
        If one of the input UngriddedData objects contains more than one
        dataset.

    Returns
    -------
    merged_stats : list
        list of `StationData` objects containing the colocated and combined
        variable data.

    """
    if add_meta_keys is None:
        add_meta_keys = []
    _check_input_data_ids_and_vars(data_ids_and_vars)
    data1, data_id1, var1 = data_ids_and_vars[0]
    data2, data_id2, var2 = data_ids_and_vars[1]

    if data2 is data1 and var2 == var1 and data_id1 == data_id2:
        raise ValueError("nothing to combine...")

    if data_id1 not in data1.contains_datasets:
        raise ValueError(f"No such data ID {data_id1} in {data1}")
    elif len(data1.contains_datasets) > 1:
        data1 = data1.extract_dataset(data_id1)

    if data_id2 not in data2.contains_datasets:
        raise ValueError(f"No such data ID {data_id2} in {data2}")
    elif len(data2.contains_datasets) > 1:
        data2 = data2.extract_dataset(data_id2)

    id1 = str(ObsVarCombi(data_id1, var1))
    id2 = str(ObsVarCombi(data_id2, var2))

    data1_stats = data1.to_station_data_all(var1, add_meta_keys=add_meta_keys)
    data1_stats["var_name"] = var1
    data1_stats["id"] = id1

    data2_stats = data2.to_station_data_all(var2, add_meta_keys=add_meta_keys)
    data2_stats["var_name"] = var2
    data2_stats["id"] = id2

    if len(data1_stats["latitude"]) <= len(data2_stats["latitude"]):  #
        short = data1_stats
        long = data2_stats
    else:
        short = data2_stats
        long = data1_stats

    match_stats_opts = ["station_name", "closest"]

    if match_stats_how not in match_stats_opts:
        raise ValueError(
            f"Invalid input for match_stats_how {match_stats_how}, choose from {match_stats_opts}"
        )

    merge_how_opts = ["combine", "mean", "eval"]

    # if e.g. merge_how is combine and var==var2, then the preferred
    # dataset & variable can be provided via this instance
    prefer = id1

    if merge_how not in merge_how_opts:
        raise ValueError(invalid_input_err_str("merge_how", merge_how, merge_how_opts))

    elif merge_how == "eval":
        if merge_eval_fun is None:
            raise ValueError("Please specify evaluation function for mode eval")
        elif not all([x in merge_eval_fun for x in [id1, id2]]):
            raise ValueError(
                f"merge_eval_fun needs to include both input "
                f"datasets;variables (e.g. {id1} + {id2}"
            )
        if "=" in merge_eval_fun:
            spl = merge_eval_fun.split("=")
            if len(spl) > 2:
                raise ValueError("merge_eval_fun contains more than 1 equality symbol...")
            var_name_out = spl[0].strip()
            merge_eval_fun = spl[1].strip()

        elif var_name_out is None:
            var_name_out = merge_eval_fun
            var_name_out = var_name_out.replace(f"{data_id1};", "")
            var_name_out = var_name_out.replace(f"{data_id2};", "")

    merge_info_vars = {"merge_how": merge_how}
    if merge_how == "combine" and var1 == var2:
        merge_info_vars["prefer"] = prefer
    elif merge_how == "eval":
        merge_info_vars["merge_eval_fun"] = merge_eval_fun

    (_index_short, _index_long, _statnames_short, _statnames_long) = _map_same_stations(
        short, long, match_stats_how, match_stats_tol_km
    )

    merged_stats = []

    var_short, var_long = short["var_name"], long["var_name"]
    for idx_short, idx_long in zip(_index_short, _index_long):
        stat_short = short["stats"][idx_short]
        stat_short.check_var_unit_aerocom(var_short)
        stat_long = long["stats"][idx_long]
        stat_long.check_var_unit_aerocom(var_long)

        # prepare output StationData object (will contain colocated timeseries
        # of both input variables as well as, additionally retrieved variable,
        # if applicable)
        new = _combine_2_sites(
            stat_short,
            var_short,
            stat_long,
            var_long,
            merge_how,
            merge_eval_fun,
            match_stats_tol_km,
            var_name_out,
            data_id_out,
            var_unit_out,
            resample_how,
            min_num_obs,
            prefer,
            merge_info_vars,
            add_meta_keys,
        )

        merged_stats.append(new)

    return merged_stats



import pytest

from pyaerocom.plot.config import _COLOR_THEMES, DEFAULT_THEME, ColorTheme, _cmap_lighttheme


def test__cmap_lighttheme():
    assert _cmap_lighttheme == "Blues"


def test__COLOR_THEMES():
    assert list(_COLOR_THEMES) == ["light", "dark"]


@pytest.mark.parametrize(
    "name, theme_name",
    [
        pytest.param(
            "bla",
            DEFAULT_THEME,
            id="invalid color theme",
            marks=pytest.mark.filterwarnings("ignore:Invalid name:UserWarning"),
        ),
        pytest.param("light", "light", id="light color theme"),
        pytest.param("dark", "dark", id="dark color theme"),
    ],
)
def test_ColorTheme___init__(name, theme_name):
    assert ColorTheme(name).name == theme_name


def test_ColorTheme_to_dict():
    value = ColorTheme("dark").to_dict()
    assert isinstance(value, dict)
    assert value == {
        "name": "dark",
        "cmap_map": "viridis",
        "cmap_map_div": "PuOr_r",
        "cmap_map_div_shifted": True,
        "color_coastline": "#e6e6e6",
        "color_map_text": "r",
    }


def test_ColorTheme_to_dict_items():
    value = ColorTheme("dark").to_dict()

    theme = ColorTheme()
    theme.from_dict(value)
    for key, val in value.items():
        assert getattr(theme, key) == val


def test_ColorTheme___str__():
    assert isinstance(str(ColorTheme("dark")), str)



from __future__ import annotations

import logging

import pytest

from pyaerocom._logging import change_verbosity


def get_level_value(logger: logging.Logger | None) -> int:
    """logging level of the first applicable StreamHandler"""

    while logger is not None:
        if not logger.hasHandlers():
            return logger.getEffectiveLevel()
        for handler in logger.handlers:
            if type(handler) is logging.StreamHandler:
                return handler.level
        if not logger.propagate:
            return logging.NOTSET
        logger = logger.parent

    return logging.NOTSET


def get_level_name(logger: logging.Logger) -> str:
    level = get_level_value(logger)
    return logging.getLevelName(level)


@pytest.fixture
def test_logger(name: str | None) -> logging.Logger:
    return logging.getLogger(name)


@pytest.mark.parametrize(
    "level", ["debug", "info", "warning", "error", "critical", 10, 20, 30, 40, 50]
)
@pytest.mark.parametrize("name", ["pyaerocom.test", "pyaerocom.deep.nested.module"])
def test_change_verbosity(level: str | int, test_logger: logging.Logger):
    change_verbosity(level)
    if isinstance(level, int):
        assert get_level_value(test_logger) == level
    if isinstance(level, str):
        assert get_level_name(test_logger) == level.upper()


@pytest.mark.parametrize(
    "level,error",
    [
        (60, "invalid logging level 60"),
        ("blaaa", "Unknown level: 'BLAAA'"),
    ],
)
@pytest.mark.parametrize("name", ["pyaerocom.test", "pyaerocom.deep.nested.module"])
def test_change_verbosity_error(level: str | int, error: str, test_logger: logging.Logger):
    with pytest.raises(ValueError) as e:
        change_verbosity(level)
    assert str(e.value).startswith(error)



import warnings

import pytest

from pyaerocom._warnings import ignore_warnings


@pytest.mark.filterwarnings("error")
def test_filter_warnings_decorator():
    @ignore_warnings(UserWarning, "User Warning")
    @ignore_warnings(DeprecationWarning, "Deprecated")
    def add_num_with_warnings(num1, num2):
        warnings.warn("User Warning", UserWarning)
        warnings.warn("Deprecated", DeprecationWarning)
        warnings.warn("General warning", UserWarning)
        return num1 + num2

    with pytest.warns(UserWarning, match="General warning"):
        add_num_with_warnings(1, 2)


@pytest.mark.filterwarnings("error")
def test_filter_warnings_category():
    """filter one warning based on the type of warning"""

    with ignore_warnings(RuntimeWarning):
        warnings.warn(RuntimeWarning("somethng unexpected..."))


@pytest.mark.filterwarnings("error")
def test_filter_warnings_message():
    """filter one warning based on the warning message"""

    with ignore_warnings(Warning, "Deprecated"):
        warnings.warn(Warning("Deprecated"))

    with ignore_warnings(Warning, "A", "B", "C"):
        warnings.warn(Warning("A"))
        warnings.warn(Warning("B"))
        warnings.warn(Warning("C"))



import logging
import os

from pyaerocom import GriddedData, TsType, const
from pyaerocom.aeroval._processing_base import DataImporter, ProcessingEngine
from pyaerocom.aeroval.modelmaps_helpers import (
    calc_contour_json,
    plot_overlay_pixel_maps,
    _jsdate_list,
    CONTOUR,
    OVERLAY,
)
from pyaerocom.colocation.colocator import Colocator

from pyaerocom.aeroval.varinfo_web import VarinfoWeb
from pyaerocom.exceptions import (
    DataCoverageError,
    DataDimensionError,
    DataQueryError,
    ModelVarNotAvailable,
    TemporalResolutionError,
    VariableDefinitionError,
    VarNotAvailableError,
    EntryNotAvailable,
)

logger = logging.getLogger(__name__)

MODELREADERS_USE_MAP_FREQ = ["ReadMscwCtm"]


class ModelMapsEngine(ProcessingEngine, DataImporter):
    """
    Engine for processing of model maps
    """

    def _get_run_kwargs(self, **kwargs):
        try:
            model_list = kwargs["model_list"]
        except KeyError:
            model_list = self.cfg.model_cfg.keylist()
        try:
            var_list = kwargs["var_list"]
        except Exception:
            var_list = None
        return model_list, var_list

    def run(self, **kwargs):
        model_list, var_list = self._get_run_kwargs(**kwargs)

        all_files = []
        for model in model_list:
            try:
                files = self._run_model(model, var_list)
            except VarNotAvailableError:
                files = []
            if not files:
                logger.warning(f"no data for model {model}, skipping")
                continue
            all_files.extend(files)
        return files

    def _get_vars_to_process(self, model_name, var_list):
        mvars = self.cfg.model_cfg.get_entry(model_name).get_vars_to_process(
            self.cfg.obs_cfg.get_all_vars()
        )[1]
        all_vars = sorted(list(set(mvars)))
        if var_list is not None:
            all_vars = [var for var in var_list if var in all_vars]
        return all_vars

    def _get_obs_vars_to_process(self, obs_name, var_list):
        ovars = self.cfg.obs_cfg.get(obs_name).get_all_vars()
        all_vars = sorted(list(set(ovars)))
        if var_list is not None:
            all_vars = [var for var in var_list if var in all_vars]
        return all_vars

    def _run_model(self, model_name: str, var_list):
        """Run evaluation of map processing

        Create json files for model-maps display. This analysis does not
        require any observation data but processes model output at all model
        grid points, which is then displayed on the website in the maps
        section.

        Parameters
        ----------
        model_name : str
            name of model to be processed
        var_list : list, optional
            name of variable to be processed. If None, all available
            observation variables are used.

        """

        try:
            var_list = self._get_vars_to_process(model_name, var_list)
        except EntryNotAvailable:
            var_list = self._get_obs_vars_to_process(model_name, var_list)

        files = []
        for var in var_list:
            logger.info(f"Processing model maps for {model_name} ({var})")

            try:  # pragma: no cover
                make_contour, make_overlay = False, False
                if isinstance(self.cfg.modelmaps_opts.plot_types, dict):
                    make_contour = CONTOUR in self.cfg.modelmaps_opts.plot_types.get(
                        model_name, False
                    )
                    make_overlay = OVERLAY in self.cfg.modelmaps_opts.plot_types.get(
                        model_name, False
                    )
                if self.cfg.modelmaps_opts.plot_types == {CONTOUR} or make_contour:
                    _files = self._process_contour_map_var(
                        model_name, var, self.reanalyse_existing
                    )
                    files.extend(_files)
                if self.cfg.modelmaps_opts.plot_types == {OVERLAY} or make_overlay:
                    # create overlay (pixel) plots
                    _files = self._process_overlay_map_var(
                        model_name, var, self.reanalyse_existing
                    )

            except ModelVarNotAvailable as ex:
                logger.warning(f"{ex}")
            except (
                TemporalResolutionError,
                DataCoverageError,
                VariableDefinitionError,
                DataQueryError,
            ) as e:
                if self.raise_exceptions:
                    raise
                logger.warning(f"Failed to process maps for {model_name} {var} data. Reason: {e}.")
        return files

    def _check_dimensions(self, data: GriddedData) -> "GriddedData":
        if not data.has_latlon_dims:
            raise DataDimensionError("data needs to have latitude an longitude dimension")
        elif not data.has_time_dim:
            raise DataDimensionError("data needs to have latitude an longitude dimension")
        if data.ndim == 4:
            data = data.extract_surface_level()
        return data

    def _process_contour_map_var(self, model_name, var, reanalyse_existing):  # pragma: no cover
        """
        Process model data to create map geojson files

        Parameters
        ----------
        model_name : str
            name of model
        var : str
            name of variable
        reanalyse_existing : bool
            if True, already existing json files will be reprocessed

        Raises
        ------
        ValueError
            If vertical code of data is invalid or not set
        AttributeError
            If the data has the incorrect number of dimensions or misses either
            of time, latitude or longitude dimension.
        ModelVarNotAvailable
            If model/var data cannot be read
        """
        try:
            data = self._read_model_data(model_name, var)
        except Exception as e:
            raise ModelVarNotAvailable(
                f"Cannot read data for model {model_name} (variable {var}): {e}"
            )

        var_ranges_defaults = self.cfg.var_scale_colmap
        if var in var_ranges_defaults.keys():
            cmapinfo = var_ranges_defaults[var]
            varinfo = VarinfoWeb(var, cmap=cmapinfo["colmap"], cmap_bins=cmapinfo["scale"])
        else:
            cmapinfo = var_ranges_defaults["default"]
            varinfo = VarinfoWeb(var, cmap=cmapinfo["colmap"], cmap_bins=cmapinfo["scale"])

        data = self._check_dimensions(data)

        outdir = self.cfg.path_manager.get_json_output_dirs()["contour"]
        outname = f"{var}_{model_name}"
        fp_geojson = os.path.join(outdir, f"{outname}.geojson")

        if not reanalyse_existing:
            if os.path.exists(fp_geojson):
                logger.info(f"Skipping contour processing of {outname}: data already exists.")
                return []

        freq = self._get_maps_freq()
        tst = TsType(data.ts_type)

        if tst < freq:
            raise TemporalResolutionError(f"need {freq} or higher, got{tst}")
        elif tst > freq:
            data = data.resample_time(str(freq))

        data.check_unit()
        # first calcualate and save geojson with contour levels
        contourjson = calc_contour_json(data, cmap=varinfo.cmap, cmap_bins=varinfo.cmap_bins)

        with self.avdb.lock():
            self.avdb.put_contour(
                contourjson,
                self.exp_output.proj_id,
                self.exp_output.exp_id,
                var,
                model_name,
            )

        return fp_geojson

    def _process_overlay_map_var(self, model_name, var, reanalyse_existing):  # pragma: no cover
        """Process overlay map (pixels) for either model or obserations
        argument model_name is a misnomer because this can also be applied to observation networks

        Args:
            model_name (str): name of model or obs to make overlay pixel maps of
            var (str): variable name
        """

        try:
            data = self.read_gridded_obsdata(model_name, var)
        except EntryNotAvailable:
            try:
                data = self._read_model_data(model_name, var)
            except Exception as e:
                raise ModelVarNotAvailable(
                    f"Cannot read data for model {model_name} (variable {var}): {e}"
                )

        var_ranges_defaults = self.cfg.var_scale_colmap

        if var in var_ranges_defaults.keys():
            cmapinfo = var_ranges_defaults[var]
            varinfo = VarinfoWeb(var, cmap=cmapinfo["colmap"], cmap_bins=cmapinfo["scale"])
        else:
            cmapinfo = var_ranges_defaults["default"]
            varinfo = VarinfoWeb(var, cmap=cmapinfo["colmap"], cmap_bins=cmapinfo["scale"])

        data = self._check_dimensions(data)

        outdir = self.cfg.path_manager.get_json_output_dirs()["contour/overlay"]

        freq = self._get_maps_freq()

        tst = TsType(data.ts_type)

        if tst < freq:
            raise TemporalResolutionError(f"need {freq} or higher, got{tst}")
        elif tst > freq:
            data = data.resample_time(str(freq))

        data.check_unit()

        tst = _jsdate_list(data)
        data = data.to_xarray()
        for i, date in enumerate(tst):
            outname = f"{model_name}_{var}_{date}"

            # Note should matche the output location defined in aerovaldb
            fp_overlay = os.path.join(outdir, outname)

            if not reanalyse_existing:
                if os.path.exists(fp_overlay):
                    logger.info(f"Skipping overlay processing of {outname}: data already exists.")
                    continue

            overlay_plot = plot_overlay_pixel_maps(
                data[i],
                cmap=varinfo.cmap,
                cmap_bins=varinfo.cmap_bins,
                format=self.cfg.modelmaps_opts.overlay_save_format,
            )

            with self.avdb.lock():
                self.avdb.put_map_overlay(
                    overlay_plot,
                    self.exp_output.proj_id,
                    self.exp_output.exp_id,
                    model_name,
                    var,
                    date,
                )

    def _get_maps_freq(self) -> TsType:
        """
        Gets the maps reading frequency. If maps_freq in cfg is coarsest, it takes the coarsest
        of the given frequencies. Else it just returns the maps_freq

        Returns
        -------
        TSType
        """
        maps_freq = TsType(self.cfg.modelmaps_opts.maps_freq)
        if maps_freq == "coarsest":  # TODO: Implement this in terms of a TsType object. #1267
            freq = min(TsType(fq) for fq in self.cfg.time_cfg.freqs)
            freq = min(freq, self.cfg.time_cfg.main_freq)
        else:
            freq = maps_freq
        return freq

    def _get_read_model_freq(self, model_ts_types: list) -> TsType:
        """
        Tries to find the best TS type to read. Checks for available ts types with the following priority

        1. If the freq from _get_maps_freq is available
        2. If maps_freq is explicitly given, and is available
        3. Iterates through the freqs given in the config, and find the coarsest available ts type

        Raises
        -------

        ValueError
            If no ts types are possible to read

        Returns
        -------
        TSType
        """
        wanted_freq = self._get_maps_freq()
        if wanted_freq in model_ts_types:
            return wanted_freq

        maps_freq = TsType(self.cfg.modelmaps_opts.maps_freq)

        if maps_freq != "coarsest":
            if maps_freq not in model_ts_types:
                raise ValueError(
                    f"Could not find any model data for given maps_freq. {maps_freq} is not in {model_ts_types}"
                )
            return maps_freq

        for freq in sorted(TsType(fq) for fq in self.cfg.time_cfg.freqs):
            if freq in model_ts_types:
                logger.info(f"Found coarsest maps_freq that is available as model data: {freq}")
                return freq

        raise ValueError("Could not find any TS type to read maps")

    def _read_model_data(self, model_name: str, var: str) -> GriddedData:
        """
        Function for reading the model data without going through the colocation object.
        This means that none of the checks normally done in the colocation class are run.

        Parameters
        ----------
        model_name : str
            name of model
        var : str
            name of variable

        Returns
        -----------
        Griddeddata
            the read data
        """
        start, stop = self.cfg.colocation_opts.start, self.cfg.colocation_opts.stop
        if self.cfg.colocation_opts.model_use_climatology:
            # overwrite start and stop to read climatology file for model
            start, stop = 9999, None

        data_id = self.cfg.model_cfg.get_entry(model_name).model_id

        try:
            data_dir = self.cfg.model_cfg.get_entry(model_name).model_data_dir
        except Exception as e:
            logger.info(f"Could not find model dir. Setting to None. Error {str(e)}")
            data_dir = None

        try:
            model_reader = self.cfg.model_cfg.get_entry(model_name).gridded_reader_id["model"]
        except Exception as e:
            logger.info(f"Could not find model reader. Setting to None. Error {str(e)}")
            model_reader = None

        if model_reader is not None:
            reader_class = Colocator.SUPPORTED_GRIDDED_READERS[model_reader]
        else:
            reader_class = Colocator.SUPPORTED_GRIDDED_READERS["ReadGridded"]

        reader = reader_class(
            data_id=data_id,
            data_dir=data_dir,
            **self.cfg.colocation_opts.model_kwargs,
        )

        if var in self.cfg.model_cfg.get_entry(model_name).model_read_aux:
            aux_instructions = self.cfg.model_cfg.get_entry(model_name).model_read_aux[var]
            reader.add_aux_compute(var_name=var, **aux_instructions)

        kwargs = {}
        kwargs.update(**self.cfg.colocation_opts.model_kwargs)
        if var in self.cfg.colocation_opts.model_read_opts:
            kwargs.update(self.cfg.colocation_opts.model_read_opts[var])
        kwargs.update(self.cfg.get_model_entry(model_name).get("model_kwargs", {}))

        if model_reader is not None and model_reader in MODELREADERS_USE_MAP_FREQ:
            ts_types = reader.ts_types
            ts_type_read = str(self._get_read_model_freq(ts_types))
        else:
            ts_type_read = self.cfg.time_cfg.main_freq

        data = reader.read_var(
            var,
            start=start,
            stop=stop,
            ts_type=ts_type_read,
            vert_which=self.cfg.colocation_opts.obs_vert_type,
            flex_ts_type=self.cfg.colocation_opts.flex_ts_type,
            **kwargs,
        )

        rm_outliers = self.cfg.colocation_opts.model_remove_outliers
        outlier_ranges = self.cfg.colocation_opts.model_outlier_ranges

        if rm_outliers:
            if var in outlier_ranges:
                low, high = outlier_ranges[var]
            else:
                var_info = const.VARS[var]
                low, high = var_info.minimum, var_info.maximum
            data.check_unit()
            data.remove_outliers(low, high, inplace=True)

        return data



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

import pytest
from xarray import load_dataarray

from pyaerocom import ColocatedData
from pyaerocom.config import ALL_REGION_NAME
from pyaerocom.griddeddata import GriddedData

from .data_access import DataForTests

CHECK_PATHS = SimpleNamespace(
    tm5="modeldata/TM5-met2010_CTRL-TEST/renamed",
    tm5aod="modeldata/TM5-met2010_CTRL-TEST/renamed/aerocom3_TM5_AP3-CTRL2016_od550aer_Column_2010_monthly.nc",
    coldata_tm5_aeronet=f"coldata/od550aer_REF-AeronetSunV3L2Subset.daily_MOD-TM5_AP3-CTRL2016_20100101_20101231_monthly_{ALL_REGION_NAME}-noMOUNTAINS.nc",
    tm5_tm5=f"coldata/od550aer_REF-TM5_AP3-CTRL2016_MOD-TM5_AP3-CTRL2016_20100101_20101231_monthly_{ALL_REGION_NAME}-noMOUNTAINS.nc",
)

TM5_DATA_PATH = DataForTests(CHECK_PATHS.tm5).path


@pytest.fixture(scope="session")
def data_tm5() -> GriddedData:
    path = DataForTests(CHECK_PATHS.tm5aod).path
    assert path.exists()
    data = GriddedData(path)
    return data


def load_coldata_tm5_aeronet_from_scratch(path: Path) -> ColocatedData:
    arr = load_dataarray(path)
    if "_min_num_obs" in arr.attrs:
        info = {}
        for val in arr.attrs["_min_num_obs"].split(";")[:-1]:
            to, fr, num = val.split(",")
            if to not in info:
                info[to] = {}
            if fr not in info[to]:
                info[to][fr] = {}
            info[to][fr] = int(num)
        arr.attrs["min_num_obs"] = info
    cd = ColocatedData()
    cd.data = arr
    return cd


@pytest.fixture(scope="session")
def coldata_tm5_aeronet() -> ColocatedData:
    path = DataForTests(CHECK_PATHS.coldata_tm5_aeronet).path
    return load_coldata_tm5_aeronet_from_scratch(path)


@pytest.fixture(scope="session")
def coldata_tm5_tm5() -> ColocatedData:
    path = DataForTests(CHECK_PATHS.tm5_tm5).path
    return load_coldata_tm5_aeronet_from_scratch(path)



"""
Interface for reading GAW files.
"""

import numpy as np
import pandas as pd

from pyaerocom import const
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData


class ReadGAW(ReadUngriddedBase):
    """Class for reading DMS data

    Extended class derived from  low-level base class :class: ReadUngriddedBase
    that contains some more functionallity.
    """

    # Mask for identifying datafiles
    _FILEMASK = "*.dat"

    # Version log of this class (for caching)
    __version__ = "0.01"

    # Name of the dataset (OBS_ID)
    DATA_ID = const.DMS_AMS_CVO_NAME  # change this since we added more vars?

    # List of all datasets supported by this interface
    SUPPORTED_DATASETS = [DATA_ID]

    # Temporal resolution flag for the supported dataset that is provided in a
    # defined temporal resolution
    TS_TYPE = "daily"

    # Dictionary specifying values corresponding to invalid measurements
    NAN_VAL = {}
    NAN_VAL["vmrdms"] = -999999999999.99
    NAN_VAL["concso4"] = -999999999999.99
    NAN_VAL["concbc"] = -999999999999.99  # Assumed, there is actually no missing data in the file
    NAN_VAL["concmsa"] = -999999999999.99
    NAN_VAL["nd"] = -9999
    NAN_VAL["sd"] = -99999.0
    NAN_VAL["f"] = -9999

    # Dictionary specifying the file column names (values) for each Aerocom
    # variable (keys)
    # Do we want CS and REM? Don't know what these columns represent...
    VAR_NAMES_FILE = {}
    VAR_NAMES_FILE["vmrdms"] = "dimethylsulfide"
    VAR_NAMES_FILE["concbc"] = "blackCarbon"
    VAR_NAMES_FILE["concso4"] = "SO4"
    VAR_NAMES_FILE["concmsa"] = "MSA"
    VAR_NAMES_FILE["nd"] = "ND"
    VAR_NAMES_FILE["sd"] = "SD"
    VAR_NAMES_FILE["f"] = "F"

    # List of variables that are provided by this dataset (will be extended
    # by auxiliary variables on class init, for details see __init__ method of
    # base class ReadUngriddedBase)
    PROVIDES_VARIABLES = list(VAR_NAMES_FILE)

    INSTRUMENT_NAME = "unknown"

    @property
    def DEFAULT_VARS(self):
        return self.PROVIDES_VARIABLES

    @property
    def DATASET_NAME(self):
        """Name of the dataset"""
        return self.data_id

    def read_file(self, filename, vars_to_retrieve=None, vars_as_series=False):
        """Read a single DMS file
        Parameters
        ----------
        filename : str
            Absolute path to filename to read.
        vars_to_retrieve : :obj:`list`, optional
            List of strings with variable names to read. If None, use :attr:
                `DEFAULT_VARS`.
        vars_as_series : bool
            If True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects.

        Returns
        -------
        StationData
            Dict-like object containing the results.

        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]

        for var in vars_to_retrieve:
            if var not in self.PROVIDES_VARIABLES:
                raise ValueError(f"Invalid input variable {var}")

        # Iterate over the lines of the file
        self.logger.info(f"Reading file {filename}")

        # We need to map the variables in the file with the values in VAR_NAME_FILE,
        # and for that we reverse the dictionary.
        # reverse dict to get keys of certain values (possible only because
        # values are hashable and unique)
        # Could also define the dict VAR_NAMES_FILE with the inverse order and
        # PROVIDES_VARIABLES = list(VAR_NAMES_FILE.values())
        inv_dict = dict(zip(self.VAR_NAMES_FILE.values(), self.VAR_NAMES_FILE.keys()))

        # Open the file, store the metadata in the first lines of the file,
        # skip empty lines, and store the headers and the data
        with open(filename) as f:
            # metadata (fixed number of rows in gaw files)
            meta = [next(f).split(":", 1)[1] for data in range(26)]

            f.readline()
            f.readline()
            f.readline()
            f.readline()
            f.readline()
            file_vars = f.readline().split()

            data = []
            for line in f:
                line = line.replace("/", "-")  # some dates have the wrong format
                rows = line.split()
                data.append(rows)

        # the next line lines should be deleted after correcting msa file
        # Drop line with empties (e.g. last line in msa file)
        for i in range(len(data)):
            if np.shape(data[i])[0] != 10:
                del data[i]

        data = np.array(data, dtype=object)

        # names of the columns in the file that I want to use
        file_vars = file_vars[5:9]
        # If we want to include CS and REM, then:
        # file_vars = file_vars[5:]
        # and add these variables to VAR_NAMES_FILE

        # The variables analized are only the intersection of those provided by
        # the dataset, vars_to_retrieve, and those available in the file,
        # file_vars.
        vars_to_retrieve_file = list(
            set(vars_to_retrieve).intersection([inv_dict[x] for x in file_vars])
        )

        # dictionary with keys and column index in the data
        data_idx = {file_vars[i]: i + 4 for i in range(0, len(file_vars))}

        # Empty data object (a dictionary with extended functionality)
        data_out = StationData()
        data_out.data_id = self.data_id
        data_out.dataset_name = self.DATASET_NAME

        # Fill dictionary with relevant metadata and variables from the file.
        # Reformat the strings, and replace whitespaces in with underscore.
        data_out["station_name"] = meta[6].strip().replace(" ", "_")
        try:
            data_out["longitude"] = float(meta[12].strip())
            data_out["latitude"] = float(meta[11].strip())
            data_out["altitude"] = float(meta[13].strip())
        except Exception as e:
            from pyaerocom.exceptions import MetaDataError

            raise MetaDataError(f"Failed to read station coordinates. {repr(e)}")
        data_out["filename"] = meta[1].strip()
        try:
            data_out["data_version"] = int(meta[5].strip())
        except Exception:
            data_out["data_version"] = None
        data_out["ts_type"] = meta[19].strip().replace(" ", "_")
        data_out["PI_email"] = meta[16].strip().replace(" ", "_")
        data_out["dataaltitude"] = meta[15].strip().replace(" ", "_")
        data_out["variables"] = vars_to_retrieve_file
        u = meta[20].strip().replace(" ", "_")

        # data in vars_to_retrieve_file
        for var in vars_to_retrieve_file:
            # find column
            print(self.VAR_NAMES_FILE[var])
            idx = data_idx[self.VAR_NAMES_FILE[var]]

            # decide what to do with values == '#REF!' or '-'. Maybe dropping the
            # entire line instead is the best option
            data[:, idx] = np.where(data[:, idx] == "#REF!", self.NAN_VAL[var], data[:, idx])
            data[:, idx] = np.where(data[:, idx] == "-", self.NAN_VAL[var], data[:, idx])

            # get data
            data_out["var_info"][var] = {}
            if idx == 4:  # variable
                if u == "ppt":
                    data_out["var_info"][var]["units"] = "mol mol-1"
                    data_out[var] = np.asarray(data[:, idx], dtype=np.float64) * 1e12
                    # reset nan values
                    data_out[var] = np.where(
                        data_out[var] == self.NAN_VAL["vmrdms"] * 1e12,
                        self.NAN_VAL["vmrdms"],
                        data_out[var],
                    )
                elif u == "ng/m3":
                    data_out["var_info"][var]["units"] = "ug/m3"
                    data_out[var] = np.asarray(data[:, idx], dtype=np.float64) * 1e-3
                    data_out[var] = np.where(
                        data_out[var] == self.NAN_VAL["concmsa"] * 1e-3,
                        self.NAN_VAL["concmsa"],
                        data_out[var],
                    )
                else:
                    data_out["var_info"][var]["units"] = u
                    data_out[var] = data[:, idx].astype(np.float64)
            else:
                data_out["var_info"][var]["units"] = 1  # If dimensionless quantity
                data_out[var] = data[:, idx].astype(np.float64)

        # If only vmrdms above, we need to write sd and f
        data_out["f"] = data[:, data_idx[self.VAR_NAMES_FILE["f"]]]
        data_out["sd"] = data[:, data_idx[self.VAR_NAMES_FILE["sd"]]]
        # Add date and time and the rest of the data to a dictionary
        data_out["dtime"] = []

        # if any hour is missing, just look at the date. This could be a
        # problem if the dataset has many observations for the same day and
        # only one observations does not specify the hour
        # Another option is to drop observations with missing hour, but then
        # a problem arises when data is not hourly.
        if any("99:99" in s for s in data[:, 1]):
            datestring = data[:, 0]
        else:
            datestring = data[:, 0] + "T" + data[:, 1]
        data_out["dtime"] = datestring.astype("datetime64[s]")

        # Replace invalid measurements with nan values
        i = 0
        for key, value in self.NAN_VAL.items():
            if key in data_out.keys():
                print(key, value, i)
                if data_out[key].dtype != "float64":
                    data_out[key] = data_out[key].astype("float64")
                data_out[key][data_out[key] == value] = np.nan
                i = i + 1

        # convert data vectors to pandas.Series (if attribute
        # vars_as_series=True)
        if vars_as_series:
            for var in vars_to_retrieve_file:
                if var in data_out:
                    data_out[var] = pd.Series(data_out[var], index=data_out["dtime"])

        return data_out

    def read(self, vars_to_retrieve=None, files=None, first_file=None, last_file=None):
        """Method that reads list of files as instance of :class:`UngriddedData`

        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or similar, optional
            List containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded.
        files : :obj:`list`, optional
            List of files to be read. If None, then the file list used is the
            returned from :func:`get_file_list`.
        first_file : :obj:`int`, optional
            Index of the first file in :obj:'file' to be read. If None, the
            very first file in the list is used.
        last_file : :obj:`int`, optional
            Index of the last file in :obj:'file' to be read. If None, the very
            last file in the list is used.

        Returns
        -------
        UngriddedData
            data object
        """

        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]

        if files is None:
            if len(self.files) == 0:
                self.get_file_list()
            files = self.files

        if first_file is None:
            first_file = 0
        if last_file is None:
            last_file = len(files)

        files = files[first_file:last_file]

        data_obj = UngriddedData()
        meta_key = 0.0
        idx = 0

        # Assign metadata object and index
        metadata = data_obj.metadata
        meta_idx = data_obj.meta_idx

        for i, _file in enumerate(files):
            station_data = self.read_file(_file, vars_to_retrieve=vars_to_retrieve)

            # only the variables in the file
            num_vars = len(station_data.var_info.keys())

            # Fill the metadata dict.
            # The location in the data set is time step dependant
            metadata[meta_key] = {}
            metadata[meta_key].update(station_data.get_meta())
            metadata[meta_key].update(station_data.get_station_coords())
            metadata[meta_key]["variables"] = list(
                station_data.var_info.keys()
            )  # vars_to_retrieve
            if "instrument_name" in station_data and station_data["instrument_name"] is not None:
                instr = station_data["instrument_name"]
            else:
                instr = self.INSTRUMENT_NAME
            metadata[meta_key]["instrument_name"] = instr

            metadata[meta_key]["var_info"] = station_data["var_info"]

            # List with indices of this station for each variable
            meta_idx[meta_key] = {}

            num_times = len(station_data["dtime"])

            totnum = num_times * num_vars

            # Check whether the size of the data object needs to be extended
            if (idx + totnum) >= data_obj._ROWNO:
                # if totnum < data_obj._CHUNKSIZE, then the latter is used
                data_obj.add_chunk(totnum)

            for var_idx, var in enumerate(list(station_data.var_info)):
                values = station_data[var]
                start = idx + var_idx * num_times
                stop = start + num_times

                # Write common meta info for this station (data lon, lat and
                # altitude are set to station locations)
                data_obj._data[start:stop, data_obj._LATINDEX] = station_data["latitude"]
                data_obj._data[start:stop, data_obj._LONINDEX] = station_data["longitude"]
                data_obj._data[start:stop, data_obj._ALTITUDEINDEX] = station_data["altitude"]
                data_obj._data[start:stop, data_obj._METADATAKEYINDEX] = meta_key
                data_obj._data[start:stop, data_obj._DATAHEIGHTINDEX] = station_data[
                    "dataaltitude"
                ]
                data_obj._data[start:stop, data_obj._DATAERRINDEX] = station_data["sd"]
                data_obj._data[start:stop, data_obj._DATAFLAGINDEX] = station_data["f"]
                data_obj._data[start:stop, data_obj._TIMEINDEX] = station_data["dtime"]
                data_obj._data[start:stop, data_obj._DATAINDEX] = values
                data_obj._data[start:stop, data_obj._VARINDEX] = var_idx
                meta_idx[meta_key][var] = np.arange(start, stop)

                if var not in data_obj.var_idx:
                    data_obj.var_idx[var] = var_idx

            idx += totnum
            meta_key = meta_key + 1.0

        # Shorten data_obj._data to the right number of points
        data_obj._data = data_obj._data[:idx]
        # data_obj.data_revision[self.DATASET_NAME] = self.data_revision

        return data_obj



import re
from pathlib import Path

import numpy as np
import pytest

from pyaerocom.io.read_aeronet_invv3 import ReadAeronetInvV3
from tests.conftest import TEST_RTOL, lustre_unavail


@lustre_unavail
@pytest.mark.xfail(raises=AssertionError)
def test_load_berlin():
    dataset = ReadAeronetInvV3()
    files = dataset.find_in_file_list("*Berlin_FUB*")
    assert len(files) == 1
    assert re.match(r"^19930101_\d{8}_Berlin_FUB.all$", Path(files[0]).name)
    data = dataset.read_file(files[0], vars_to_retrieve=["abs550aer"])

    test_vars = ["abs440aer", "angabs4487aer", "abs550aer"]
    assert all(x in data for x in test_vars)

    # more than 100 timestamps
    assert all(len(data[x]) > 100 for x in test_vars)

    assert isinstance(data["dtime"][0], np.datetime64)
    t0 = data["dtime"][0]

    assert t0 == np.datetime64("2014-07-07T12:00:00")

    first_vals = [np.nanmean(data[var]) for var in test_vars]

    nominal = [0.015538, 0.915505, 0.012879]
    assert first_vals == pytest.approx(nominal, rel=TEST_RTOL)



from __future__ import annotations

import sys

from pyaerocom.data import resources

if sys.version_info >= (3, 11):  # pragma: no cover
    import tomllib
else:  # pragma: no cover
    import tomli as tomllib

_META_KEYS = "meta_keys.toml"


def ghost_meta_keys() -> list[str]:
    assert resources.is_resource(__package__, _META_KEYS), f"{_META_KEYS} missing in {__package__}"
    variables = tomllib.loads(resources.read_text(__package__, _META_KEYS))
    return variables["ghost_meta_keys"]



import xarray
from pyproj import CRS, transformer


class ProjectionInformationException(Exception):
    pass


class ProjectionInformation:
    def __init__(self):
        self._crs = None
        self._x_axis = None
        self._y_axis = None
        self._units = None

    @property
    def x_axis(self):
        return self._x_axis

    @property
    def y_axis(self):
        return self._y_axis

    def to_proj(self, latitude, longitude):
        """convert latitude and longitude to x and y

        :param latitude: latitude values
        :param longitude: longitude values
        :return: tuple of x and y coordinates
        """
        trans = transformer.Transformer.from_crs(4326, self._crs)
        return trans.transform(latitude, longitude)

    def to_latlon(self, x, y):
        """convert x and y to latitude and longitude

        :param x: x values
        :param y: y values
        :return: tuple of latitude and longitude coordinates
        """
        trans = transformer.Transformer.from_crs(self._crs, 4326)
        return trans.transform(x, y)

    @staticmethod
    def from_xarray(ds: xarray.Dataset, var: str):
        """initialize the ProjectionInformation from an xarray variable

        returns None if no projection exists for the variable or ProjectionInformation
        """
        da = ds[var]
        if "grid_mapping" not in da.attrs:
            return None
        pi = ProjectionInformation()
        pi._crs = CRS.from_cf(ds[da.grid_mapping].attrs)

        for c in da.coords:
            if len(da.coords[c].dims) != 1:
                continue
            if "standard_name" in da.coords[c].attrs:
                if da.coords[c].standard_name in (
                    "longitude",
                    "grid_longitude",
                    "projection_x_coordinate",
                ):
                    pi._x_axis = c
                    break
            if "axis" in da.coords[c].attrs:
                if da.coords[c].axis in ("x", "X"):
                    pi._x_axis = c
                    break
        for c in da.coords:
            if len(da.coords[c].dims) != 1:
                continue
            if "standard_name" in da.coords[c].attrs:
                if da.coords[c].standard_name in (
                    "latitude",
                    "grid_latitude",
                    "projection_y_coordinate",
                ):
                    pi._y_axis = c
                    break
            if "axis" in da.coords[c].attrs:
                if da.coords[c].axis in ("y", "Y"):
                    pi._y_axis = c
                    break
        if pi._x_axis is None or pi._y_axis is None:
            raise ProjectionInformationException(f"no x or y axis found for variable '{var}'")
        pi._units = da.coords[pi._y_axis].units
        return pi



from pyaerocom.io.mscw_ctm.model_variables import emep_variables


def test_emep_variables():
    variables = emep_variables()
    assert isinstance(variables, dict)
    assert variables["conco3"] == "SURF_ug_O3"



import importlib
import os
import sys
from collections.abc import Callable

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from typing import TYPE_CHECKING

from pydantic import (
    BaseModel,
    model_validator,
)

from pyaerocom._lowlevel_helpers import AsciiFileLoc


def check_aux_info(fun, vars_required, funcs):
    """
    Make sure information is correct for computation of auxiliary variables

    Parameters
    ----------
    fun : str or callable
        name of function or function used to compute auxiliary variable. If
        str, then arg `funcs` needs to be provided.
    vars_required : list
        list of required variables for computation of auxiliary variable.
    funcs : dict
        Dictionary with possible functions (values) and names (keys)

    Returns
    -------
    dict
        dict containing callable function object and list of variables
        required.

    """
    spec = _AuxReadSpec(fun=fun, vars_required=vars_required, funcs=funcs)
    return dict(fun=spec.fun, vars_required=spec.vars_required)


class _AuxReadSpec(BaseModel):
    """
    Class that specifies requirements for computation of additional variables

    Attributes
    ----------
    vars_required : list
        list of required variables for computation of auxiliary variable.
    fun : callable
        function used to compute auxiliary variable.

    Parameters
    ----------
    fun : str or callable
        name of function or function used to compute auxiliary variable. If
        str, then arg `funcs` needs to be provided.
    vars_required : list
        list of required variables for computation of auxiliary variable.
    funcs : dict
        Dictionary with possible functions (values) and names (keys)

    """

    if TYPE_CHECKING:
        fun: Callable
    else:
        fun: str | Callable
    vars_required: list[str]
    funcs: dict[str, Callable]

    @model_validator(mode="after")
    def validate_fun(self) -> Self:
        if callable(self.fun):
            return self
        elif isinstance(self.fun, str):
            self.fun = self.funcs[self.fun]
            return self
        else:
            raise ValueError("failed to retrieve aux func")


class ReadAuxHandler:
    """
    Helper class for import of auxiliary function objects

    Attributes
    ----------
    aux_file : str
        path to python module containing function definitions (note: function
        definitions in module need to be stored in a dictionary called
        `FUNS` in the file, where keys are names of the functions and
        values are callable objects.)

    Parameters
    ----------
    aux_file : str
        input file containing auxiliary functions (details see Attributes
        section).
    """

    aux_file = AsciiFileLoc(assert_exists=True, auto_create=False)

    def __init__(self, aux_file: str):
        self.aux_file = aux_file

    def import_module(self):
        """
        Import :attr:`aux_file` as python module

        Uses :func:`importlib.import_module` for import.
        Returns
        -------
        module
            imported module.

        """
        moddir, fname = os.path.split(self.aux_file)
        if moddir not in sys.path:
            sys.path.append(moddir)
        modname = fname.split(".")[0]
        return importlib.import_module(modname)

    def import_all(self):
        """
        Import all callable functions in module with their names

        Currently, these are expected to be stored in a dictionary called
        `FUNS` which should be defined in the python module.

        Returns
        -------
        dict
            function definitions.

        """
        mod = self.import_module()
        return mod.FUNS



from __future__ import annotations

import pandas as pd
import pytest

from pyaerocom.exceptions import UnitConversionError
from pyaerocom.units_helpers import (
    _check_unit_endswith_freq,
    _unit_conversion_fac_custom,
    _unit_conversion_fac_si,
    convert_unit,
    get_unit_conversion_fac,
)


@pytest.mark.parametrize(
    "unit,result",
    [
        ("ug min-1", True),
        ("ug/min", True),
        ("ug", False),
        ("ug m-1", False),
        ("ug m-2", False),
        ("kg m-2/yr", True),
        ("kg m-2/month", True),
        ("kg m-2/week", True),
        ("kg m-2 month-1", True),
        ("mm", False),
        ("mg d-1", True),
    ],
)
def test__check_unit_endswith_freq(unit: str, result: bool):
    assert _check_unit_endswith_freq(unit) == result


@pytest.mark.parametrize(
    "from_unit,to_unit,result",
    [
        ("m-1", "1/Mm", 1e6),
        ("ug m-3", "ug/m3", 1),
    ],
)
def test__unit_conversion_fac_si(from_unit: str, to_unit: str, result: float):
    assert _unit_conversion_fac_si(from_unit, to_unit) == result


@pytest.mark.parametrize(
    "var_name,from_unit,to_unit,result,",
    [
        # ("SO4ugSm3", "ug/m3", "ug S m-3", 0.333798316),
        # ("concno3", "ug N m-3", "ug m-3", 4.426717),
        ("concso2", "ug S/m3", "ug m-3", 1.9979),
        # ("concso4", "ug S/m3", "ug m-3", 2.9958),
        ("concbc", "ug C/m3", "ug m-3", 1),
        ("concoa", "ug C/m3", "ug m-3", 1),
        ("concoc", "ug C/m3", "ug m-3", 1),
        ("concpm25", "ug m-3", "1", 1),
        ("concpm10", "ug m-3", "1", 1),
        ("wetso4", "kg S/ha", "kg m-2", 0.0003),
        ("concso4pr", "mg S/L", "g m-3", 2.995821),
    ],
)
def test__unit_conversion_fac_custom(var_name: str, from_unit: str, to_unit: str, result: float):
    converted_unit, conversion_fac = _unit_conversion_fac_custom(var_name, from_unit)
    assert converted_unit == to_unit
    assert conversion_fac == pytest.approx(result, rel=1e-2)


def test__unit_conversion_fac_custom_error():
    with pytest.raises(UnitConversionError) as e:
        _unit_conversion_fac_custom("concNno3", "ug N m-3")
    assert str(e.value).startswith("Failed to convert unit ug N/m3 (variable concNno3).")


def test__unit_conversion_fac_custom_FAIL(monkeypatch):
    MOCK_UCONV_MUL_FACS = pd.DataFrame(
        [
            ["concso4", "ug S/m3", "ug m-3", 1],
            ["concso4", "ug S/m3", "ug m-3", 2],
        ],
        columns=["var_name", "from", "to", "fac"],
    ).set_index(["var_name", "from"])
    monkeypatch.setattr("pyaerocom.units_helpers.UCONV_MUL_FACS", MOCK_UCONV_MUL_FACS)

    with pytest.raises(UnitConversionError) as e:
        _unit_conversion_fac_custom("concso4", "ug S/m3")
    assert "Could not find unique conversion factor in table" in str(e.value)


@pytest.mark.parametrize(
    "from_unit,to_unit,var_name,result",
    [
        ("ug m-3", "ug/m3", None, 1),
        ("ug m-3", "ug/m3", "concso2", 1),
        ("mg m-3", "ug m-3", "concso2", 1e3),
        ("ug S/m3", "mg m-3", "concso2", 1.9979e-3),
    ],
)
def test_convert_unit(from_unit: str, to_unit: str, var_name: str, result: float):
    converted = convert_unit(1, from_unit, to_unit, var_name)
    assert converted == pytest.approx(result, rel=1e-2)


@pytest.mark.parametrize(
    "from_unit,to_unit,var_name,ts_type,result",
    [
        ("kg (2m)-2", "kg/m2", None, None, 0.25),
        ("mm", "mm d-1", "prmm", "daily", 1),
        ("mm", "mm d-1", "prmm", "hourly", 24),
        ("mg m-2", "ug m-2 d-1", "wetoxs", "hourly", 24e3),
        ("mg m-2", "mg m-2 d-1", "wetoxs", "hourly", 24),
        ("mg m-2", "mg m-2/d", "wetoxs", "daily", 1),
        ("mg m-2", "mg m-2 d-1", "wetoxs", "daily", 1),
        ("mg m-2", "ug mm-2", None, None, 1e-3),
        ("mg", "ug", None, None, 1000),
        ("1", "1", None, None, 1),
    ],
)
def test_get_unit_conversion_fac(
    from_unit: str, to_unit: str, var_name: str | None, ts_type: str | None, result: float
):
    conversion_fac = get_unit_conversion_fac(from_unit, to_unit, var_name, ts_type)
    assert conversion_fac == pytest.approx(result, rel=1e-3)


@pytest.mark.parametrize(
    "from_unit,to_unit,var_name",
    [
        ("kg m-2", "mm", "pr"),
        ("kg m-2", "mm", "od550aer"),
        ("1", "ug", None),
    ],
)
def test_get_unit_conversion_fac_error(from_unit: str, to_unit: str, var_name: str | None):
    with pytest.raises(UnitConversionError) as e:
        get_unit_conversion_fac(from_unit, to_unit, var_name)
    assert str(e.value) == f"failed to convert unit from {from_unit} to {to_unit}"



"""
High level I/O utility methods for pyaerocom
"""

import pytest

from pyaerocom.io import utils
from pyaerocom.io.read_aasetal import ReadAasEtal
from pyaerocom.io.read_aeronet_invv3 import ReadAeronetInvV3
from pyaerocom.io.read_aeronet_sdav3 import ReadAeronetSdaV3
from pyaerocom.io.read_aeronet_sunv3 import ReadAeronetSunV3
from pyaerocom.io.read_ebas import ReadEbas
from tests.conftest import lustre_unavail


def name(obj):
    return obj.__name__


TESTDATA = [
    ("EBASMC", ReadEbas),
    ("AeronetSunV3Lev2.daily", ReadAeronetSunV3),
    ("AeronetSDAV3Lev2.daily", ReadAeronetSdaV3),
    ("AeronetInvV3Lev2.daily", ReadAeronetInvV3),
    ("GAWTADsubsetAasEtAl", ReadAasEtal),
]


@pytest.mark.parametrize(("obs_id,reader"), TESTDATA)
def test_get_ungridded_reader(obs_id, reader):
    assert name(utils.get_ungridded_reader(obs_id)) == name(reader)


@lustre_unavail
def test_browse_database():
    assert "TM5-met2010_AP3-CTRL2019" in utils.browse_database("*TM5*CTRL*")



import pytest

import pyaerocom._concprcp_units_helpers


@pytest.mark.parametrize(
    "unit,ts_type,result",
    [
        pytest.param("mg m-2/h", "hourly", "mg m-2 h-1", id="mg/m2/h"),
        pytest.param("mg N m-2", "hourly", "mg N m-2 h-1", id="mg(N)/m2/h"),
        pytest.param("mg N m-2", "daily", "mg N m-2 d-1", id="mg(N)/m2/day"),
    ],
)
def test_translate_rate_units_implicit(unit, ts_type, result):
    val = pyaerocom._concprcp_units_helpers.translate_rate_units_implicit(unit, ts_type)
    assert val == result



from __future__ import annotations

import pytest

from pyaerocom.vertical_profile import VerticalProfile


@pytest.fixture(scope="module")
def vertical_profile() -> VerticalProfile:
    return VerticalProfile(
        data=[1, 4, 3, 2, 1],
        altitude=[1, 2, 3, 4, 5],
        dtime=[1],
        var_name="bla",
        data_err=[1, 1, 1, 1, 1],
        var_unit="ug m-3",
        altitude_unit="m",
    )


@pytest.mark.parametrize(
    "kwargs,exception,error",
    [
        pytest.param(
            dict(),
            TypeError,
            "missing 7 required positional arguments: 'data', 'altitude', 'dtime', 'var_name', 'data_err', 'var_unit', and 'altitude_unit'",
            id="no args",
        ),
        pytest.param(
            dict(
                data=[1, 2],
                altitude=[1],
                dtime=[1],
                var_name="bla",
                data_err=[1],
                var_unit="1",
                altitude_unit="1",
            ),
            AssertionError,
            "",
            id="wrong data",
        ),
    ],
)
def test_VerticalProfile_error(kwargs: dict, exception: type[Exception], error: str):
    with pytest.raises(exception) as e:
        VerticalProfile(**kwargs)
    assert str(e.value).endswith(error)


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(),
        dict(plot_errs=False),
        dict(whole_alt_range=True),
        dict(errs_shaded=False),
        dict(add_vertbar_zero=False),
    ],
)
def test_VerticalProfile_plot(vertical_profile: VerticalProfile, kwargs: dict[str, bool]):
    vertical_profile.plot(**kwargs)



import logging
from time import time

from cf_units import Unit
from numpy.typing import ArrayLike

from pyaerocom import ColocatedData, TsType
from pyaerocom.aeroval._processing_base import ProcessingEngine
from pyaerocom.aeroval.coldatatojson_helpers import (
    _apply_annual_constraint,
    _init_data_default_frequencies,
    _init_meta_glob,
    _process_heatmap_data,
    _process_map_and_scat,
    _process_regional_timeseries,
    _process_sites,
    _process_sites_weekly_ts,
    _process_statistics_timeseries,
    _remove_less_covered,
    init_regions_web,
    process_profile_data_for_regions,
    process_profile_data_for_stations,
)
from pyaerocom.aeroval.exceptions import ConfigError
from pyaerocom.aeroval.json_utils import round_floats
from pyaerocom.exceptions import TemporalResolutionError

logger = logging.getLogger(__name__)


class ColdataToJsonEngine(ProcessingEngine):
    def run(self, files):
        """
        Convert colocated data files to json

        Parameters
        ----------
        files : list
            list of file paths pointing to colocated data objects to be
            processed.

        Returns
        -------
        list
            list of files that have been converted.

        """

        converted = []
        for file in files:
            logger.info(f"Processing: {file}")
            coldata = ColocatedData(data=file)
            self.process_coldata(coldata)
            converted.append(file)
        return converted

    def process_coldata(self, coldata: ColocatedData):
        """
        Creates all json files for one ColocatedData object

        Parameters
        ----------
        coldata : ColocatedData
            colocated data to be processed.

        Raises
        ------
        NotImplementedError
            DESCRIPTION.
        ValueError
            DESCRIPTION.
        ConfigError
            DESCRIPTION.

        Returns
        -------
        None.

        """
        t00 = time()
        use_weights = self.cfg.statistics_opts.weighted_stats
        drop_stats = self.cfg.statistics_opts.drop_stats
        # redundant, but cheap and important to be correct
        self.cfg._check_time_config()
        freqs = self.cfg.time_cfg.freqs
        periods = self.cfg.time_cfg.periods
        seasons = self.cfg.time_cfg.get_seasons()
        main_freq = self.cfg.time_cfg.main_freq
        annual_stats_constrained = self.cfg.statistics_opts.annual_stats_constrained

        out_dirs = self.cfg.path_manager.get_json_output_dirs(True)
        regions_how = self.cfg.webdisp_opts.regions_how

        stats_min_num = self.cfg.statistics_opts.MIN_NUM

        if hasattr(coldata.data, "altitude_units"):
            if Unit(coldata.data.attrs["altitude_units"]) != Unit(
                "km"
            ):  # put everything in terms of km for viz
                # convert start and end for file naming
                self._convert_coldata_altitude_units_to_km(coldata)

        vert_code = self._get_vert_code(coldata)

        diurnal_only = coldata.get_meta_item("diurnal_only")

        add_trends = self.cfg.statistics_opts.add_trends
        trends_min_yrs = self.cfg.statistics_opts.stats_min_yrs

        min_yrs = self.cfg.statistics_opts.obs_min_yrs
        sequential_yrs = self.cfg.statistics_opts.sequential_yrs
        avg_over_trends = self.cfg.statistics_opts.avg_over_trends
        use_fairmode = self.cfg.statistics_opts.use_fairmode
        use_diurnal = self.cfg.statistics_opts.use_diurnal

        # ToDo: some of the checks below could be done automatically in
        # EvalSetup, and at an earlier stage
        if vert_code == "ModelLevel":
            raise NotImplementedError("Coming (not so) soon...")

        # this will need to be figured out as soon as there is altitude
        elif "altitude" in coldata.data.dims:
            raise ValueError("Altitude should have been dealt with already in the colocation")

        elif not isinstance(coldata, ColocatedData):
            raise ValueError(f"Need ColocatedData object, got {type(coldata)}")

        elif coldata.has_latlon_dims and regions_how == "country":
            raise NotImplementedError(
                "Cannot yet apply country filtering for 4D colocated data instances"
            )
        elif main_freq not in freqs:
            raise ConfigError(f"main_freq {main_freq} is not in experiment frequencies: {freqs}")
        if self.cfg.statistics_opts.stats_tseries_base_freq is not None:
            if self.cfg.statistics_opts.stats_tseries_base_freq not in freqs:
                raise ConfigError(
                    f"Base frequency for statistics timeseries needs to be "
                    f"specified in experiment frequencies: {freqs}"
                )
        # init some stuff
        if "var_name_input" in coldata.metadata:
            obs_var = coldata.metadata["var_name_input"][0]
            model_var = coldata.metadata["var_name_input"][1]
        elif (
            "obs_vars" in coldata.metadata
        ):  # Try and get from obs_vars. Should not be needed in reading in a ColocatedData object created with pyaerocom.
            obs_var = model_var = coldata.metadata["obs_vars"]
            coldata.metadata["var_name_input"] = [obs_var, model_var]
            logger.warning(
                "ColdataToJsonEngine: Failed to access var_name_input from coldata.metadata. "
                "This could be because you're using a ColocatedData object created outside of pyaerocom. "
                "Setting obs_var and model_data to value in obs_vars instead. "
                "Setting var_input_data to these values as well. "
            )

        else:
            obs_var = model_var = "UNDEFINED"

        model_name = coldata.model_name
        obs_name = coldata.obs_name

        mcfg = self.cfg.model_cfg.get_entry(model_name)
        var_name_web = mcfg.get_varname_web(model_var, obs_var)

        logger.info(
            f"Computing json files for {model_name} ({model_var}) vs. {obs_name} ({obs_var})"
        )

        meta_glob = _init_meta_glob(
            coldata,
            vert_code=vert_code,
            obs_name=obs_name,
            model_name=model_name,
            var_name_web=var_name_web,
        )
        if min_yrs > 0:
            logger.info(
                f"Removing stations with less than {min_yrs} years of data, with sequential_yrs = {sequential_yrs}"
            )
            coldata = _remove_less_covered(coldata, min_yrs, sequential_yrs)

        # get region IDs
        (regborders, regs, regnames) = init_regions_web(coldata, regions_how)

        # Synchronise regions.json file
        with self.avdb.lock():
            regions = self.avdb.get_regions(
                self.exp_output.proj_id, self.exp_output.exp_id, default={}
            )
            for region_name, region_info in regborders.items():
                regions[region_name] = round_floats(region_info)
            self.avdb.put_regions(regions, self.exp_output.proj_id, self.exp_output.exp_id)

        use_country = True if regions_how == "country" else False

        data = _init_data_default_frequencies(coldata, freqs)

        if annual_stats_constrained:
            data = _apply_annual_constraint(data)

        if not coldata.data.attrs.get("just_for_viz", False):  # make the regular json output
            if not diurnal_only:
                logger.info("Processing statistics timeseries for all regions")

                self._process_stats_timeseries_for_all_regions(
                    data=data,
                    coldata=coldata,
                    main_freq=main_freq,
                    regnames=regnames,
                    use_weights=use_weights,
                    drop_stats=drop_stats,
                    use_country=use_country,
                    obs_name=obs_name,
                    obs_var=obs_var,
                    var_name_web=var_name_web,
                    out_dirs=out_dirs,
                    vert_code=vert_code,
                    model_name=model_name,
                    model_var=model_var,
                    meta_glob=meta_glob,
                    periods=periods,
                    seasons=seasons,
                    add_trends=add_trends,
                    trends_min_yrs=trends_min_yrs,
                    regions_how=regions_how,
                    regs=regs,
                    stats_min_num=stats_min_num,
                    use_fairmode=use_fairmode,
                    avg_over_trends=avg_over_trends,
                )
            if coldata.ts_type == "hourly" and use_diurnal:
                logger.info("Processing diurnal profiles")
                self._process_diurnal_profiles(
                    coldata=coldata,
                    regions_how=regions_how,
                    regnames=regnames,
                    meta_glob=meta_glob,
                )
        else:
            logger.info("Processing profile data for visualization")

            self._process_profile_data_for_visualization(
                data=data,
                use_country=use_country,
                region_names=regnames,
                station_names=coldata.data.station_name.values,
                periods=periods,
                seasons=seasons,
                obs_name=obs_name,
                var_name_web=var_name_web,
            )

        logger.info(
            f"Finished computing json files for {model_name} ({model_var}) vs. "
            f"{obs_name} ({obs_var})"
        )

        dt = time() - t00
        logger.info(f"Time expired: {dt:.2f} s")

    def _convert_coldata_altitude_units_to_km(self, coldata: ColocatedData = None):
        alt_units = coldata.data.attrs["altitude_units"]

        coldata.data.attrs["vertical_layer"]["start"] = str(
            Unit(alt_units).convert(coldata.data.attrs["vertical_layer"]["start"], other="km")
        )

        coldata.data.attrs["vertical_layer"]["end"] = str(
            Unit(alt_units).convert(coldata.data.attrs["vertical_layer"]["end"], other="km")
        )

    def _get_vert_code(self, coldata: ColocatedData = None):
        if hasattr(coldata.data, "vertical_layer"):
            # start and end for vertical layers (display on web and name jsons)
            start = float(coldata.data.attrs["vertical_layer"]["start"])
            end = float(coldata.data.attrs["vertical_layer"]["end"])
            # format correctly (e.g., 1, 1.5, 2, 2.5, etc.)
            start = f"{round(float(start), 1):g}"
            end = f"{round(float(end), 1):g}"
            vert_code = f"{start}-{end}km"
        else:
            vert_code = coldata.get_meta_item("vert_code")
        return vert_code

    def _process_profile_data_for_visualization(
        self,
        data: dict[str, ColocatedData] = None,
        use_country: bool = False,
        region_names: dict[str:str] = None,
        station_names: ArrayLike = None,
        periods: tuple[str, ...] = None,
        seasons: tuple[str, ...] = None,
        obs_name: str = None,
        var_name_web: str = None,
    ):
        if region_names is None and station_names is None:
            raise ValueError("Both region_id and station_name can not both be None")

        # Loop through regions
        for regid in region_names:
            profile_viz = process_profile_data_for_regions(
                data=data,
                region_id=regid,
                use_country=use_country,
                periods=periods,
                seasons=seasons,
            )
            location = region_names[regid]
            self.exp_output.add_profile_entry(
                data,
                profile_viz,
                periods,
                seasons,
                location=location,
                network=obs_name,
                obsvar=var_name_web,
            )

        # Loop through stations
        for station_name in station_names:
            profile_viz = process_profile_data_for_stations(
                data=data,
                station_name=station_name,
                use_country=use_country,
                periods=periods,
                seasons=seasons,
            )

            self.exp_output.add_profile_entry(
                data,
                profile_viz,
                periods,
                seasons,
                location=station_name,
                network=obs_name,
                obsvar=var_name_web,
            )

    def _process_stats_timeseries_for_all_regions(
        self,
        data: dict[str, ColocatedData] = None,
        coldata: ColocatedData = None,
        main_freq: str = None,
        regnames=None,
        use_weights: bool = True,
        drop_stats: tuple = (),
        use_country: bool = False,
        obs_name: str = None,
        obs_var: str = None,
        var_name_web: str = None,
        out_dirs: dict = None,
        vert_code: str = None,
        model_name: str = None,
        model_var: str = None,
        meta_glob: dict = None,
        periods: tuple[str, ...] = None,
        seasons: tuple[str, ...] = None,
        add_trends: bool = False,
        trends_min_yrs: int = 7,
        regions_how: str = "default",
        regs: dict = None,
        stats_min_num: int = 1,
        use_fairmode: bool = False,
        avg_over_trends: bool = False,
    ):
        input_freq = self.cfg.statistics_opts.stats_tseries_base_freq
        for reg in regnames:
            try:
                stats_ts = _process_statistics_timeseries(
                    data=data,
                    freq=main_freq,
                    region_ids={reg: regnames[reg]},
                    use_weights=use_weights,
                    drop_stats=drop_stats,
                    use_country=use_country,
                    data_freq=input_freq,
                )

            except TemporalResolutionError:
                stats_ts = {}

            region = regnames[reg]
            self.exp_output.add_heatmap_timeseries_entry(
                stats_ts,
                region,
                obs_name,
                var_name_web,
                vert_code,
                model_name,
                model_var,
            )

        logger.info("Processing heatmap data for all regions")

        hm_all = _process_heatmap_data(
            data,
            regnames,
            use_weights,
            drop_stats,
            use_country,
            meta_glob,
            periods,
            seasons,
            add_trends,
            trends_min_yrs,
            avg_over_trends,
        )

        for freq, hm_data in hm_all.items():
            self.exp_output.add_heatmap_entry(
                hm_data, freq, obs_name, var_name_web, vert_code, model_name, model_var
            )

        logger.info("Processing regional timeseries for all regions")
        ts_objs_regional = _process_regional_timeseries(data, regnames, regions_how, meta_glob)

        self.exp_output.write_timeseries(ts_objs_regional)
        if coldata.has_latlon_dims:
            for cd in data.values():
                if cd is not None:
                    cd.data = cd.flatten_latlondim_station_name().data

        logger.info("Processing individual site timeseries data")
        (ts_objs, map_meta, site_indices) = _process_sites(data, regs, regions_how, meta_glob)

        self.exp_output.write_timeseries(ts_objs)

        scatter_freq = min(TsType(fq) for fq in self.cfg.time_cfg.freqs)
        scatter_freq = min(scatter_freq, main_freq)

        logger.info("Processing map and scat data by period")

        for period in periods:
            # compute map_data and scat_data just for this period
            map_data, scat_data = _process_map_and_scat(
                data,
                map_meta,
                site_indices,
                [period],
                str(scatter_freq),
                stats_min_num,
                seasons,
                add_trends,
                trends_min_yrs,
                avg_over_trends,
                use_fairmode,
                obs_var,
                drop_stats,
            )

            with self.avdb.lock():
                self.avdb.put_map(
                    map_data,
                    self.exp_output.proj_id,
                    self.exp_output.exp_id,
                    obs_name,
                    var_name_web,
                    vert_code,
                    model_name,
                    model_var,
                    period.replace("/", ""),  # Remove slashes in CAMS2_83 period.
                )

                self.avdb.put_scatter(
                    scat_data,
                    self.exp_output.proj_id,
                    self.exp_output.exp_id,
                    obs_name,
                    var_name_web,
                    vert_code,
                    model_name,
                    model_var,
                    period.replace("/", ""),  # Remove slashes in CAMS2_83 period.
                )

    def _process_diurnal_profiles(
        self,
        coldata: ColocatedData = None,
        regions_how: str = "default",
        regnames=None,
        meta_glob: dict = None,
    ):
        ts_objs_weekly, ts_objs_weekly_reg = _process_sites_weekly_ts(
            coldata, regions_how, regnames, meta_glob
        )

        for ts_data_weekly in ts_objs_weekly:
            self.exp_output.write_station_data(ts_data_weekly)
        if ts_objs_weekly_reg is not None:
            for ts_data_weekly_reg in ts_objs_weekly_reg:
                self.exp_output.write_station_data(ts_data_weekly_reg)



"""
Global config for ratio pm10 vs pm25
"""

import copy
import logging
import os

logger = logging.getLogger(__name__)

# Constraints
DEFAULT_RESAMPLE_CONSTRAINTS = dict(
    yearly=dict(monthly=9),
    monthly=dict(
        daily=21,
        weekly=3,
    ),
    daily=dict(hourly=18),
)

DEFAULT_RESAMPLE_CONSTRAINTS_DAILY = dict(
    daily=dict(hourly=18),
)

# ODCSFUN_EEANRT = "EEAAQeRep.NRT;concpm10/EEAAQeRep.NRT;concpm25"
ODCSFUN_EEAV2 = "EEAAQeRep.v2;concpm10/EEAAQeRep.v2;concpm25"
ODCSFUN_EBAS = "EBASMC;concpm10/EBASMC;concpm25"


def get_CFG(reportyear, year, model_dir) -> dict:
    """create aeroval configuration dict to run the variable
    ratpm10pm25 (ratio pm10 vspm25)

    :returns: a dict of a model configuration usable for EvalSetup
    """
    # get current path for reference to local gridded_io_aux.py

    CFG = dict(
        json_basedir=os.path.abspath("/home/jang/data/aeroval-local-web/data"),
        coldata_basedir=os.path.abspath("/home/jang/data/aeroval-local-web/coldata"),
        # io_aux_file=os.path.abspath("/home/jang/data/aeroval-local-web/gridded_io_aux.py"), not needed for ReadMscwCtm
        # io_aux_file=os.path.join(base_conf_path, "gridded_io_aux.py"),
        # var_scale_colmap_file=os.path.abspath(
        #     "/home/jang/data/aeroval-local-web/pyaerocom-config/config_files/CAMEO/user_var_scale_colmap.ini"
        # ),
        # if True, existing colocated data files will be deleted and contours will be overwritten
        reanalyse_existing=True,
        only_json=False,
        add_model_maps=False,
        only_model_maps=False,
        modelmaps_opts=dict(maps_freq="monthly", maps_res_deg=5),
        clear_existing_json=False,
        # if True, the analysis will stop whenever an error occurs (else, errors that
        # occurred will be written into the logfiles)
        raise_exceptions=False,
        # Regional filter for analysis
        filter_name="ALL-wMOUNTAINS",
        # colocation frequency (no statistics in higher resolution can be computed)
        ts_type="daily",
        map_zoom="Europe",
        freqs=[
            "yearly",
            "monthly",
            "daily",
        ],
        periods=[f"{year}"],
        main_freq="monthly",
        zeros_to_nan=False,
        use_diurnal=True,
        min_num_obs=DEFAULT_RESAMPLE_CONSTRAINTS,
        colocate_time=True,
        obs_remove_outliers=False,
        model_remove_outliers=False,
        harmonise_units=True,
        regions_how="country",
        annual_stats_constrained=True,
        proj_id="emep",
        exp_id=f"{reportyear}-reporting",
        exp_name=f"Evaluation of EMEP runs for {reportyear} EMEP reporting",
        exp_descr=(
            f"Evaluation of EMEP runs for {reportyear} EMEP reporting. The EMEP model, simulated for {year}, is compared against observations from EEA and EBAS."
        ),
        exp_pi="emep.mscw@met.no",
        public=True,
        # directory where colocated data files are supposed to be stored
        weighted_stats=True,
        var_order_menu=[
            # Gases
            "ratpm10pm25",
            "ratpm25pm10",
            "concNno",
            "concNno2",
            "concNtno3",
            "concNhno3",
            "concNtnh",
            "concNnh3",
            "concnh4",
            "concSso2",
            "concso4t",
            "concso4c",
            "vmro3",
            "vmro3max",
            "vmro3mda8",
            "vmrox",
            "vmrco",
            # PMs
            "concpm10",
            "concpm25",
            "concno3pm10",
            "concno3pm25",
            "concnh4pm25",
            "concso4pm25",
            "concCecpm10",
            "concCecpm25",
            "concCocpm10",  # SURF_ugC_PM_OMCOARSE missing in model-output
            "concCocpm25",
            "concsspm10",
            "concsspm25",
            # Depositions
            "wetrdn",
            "wetoxs",
            "wetoxn",
            "prmm",
        ],
    )

    CFG["model_cfg"] = {
        "EMEPcameo": dict(
            model_id="EMEP,",
            model_data_dir=model_dir,
            gridded_reader_id={"model": "ReadMscwCtm"},
            # model_read_aux={},
            model_ts_type_read="daily",
        ),
    }

    """
    Filters
    """

    BASE_FILTER = {
        "latitude": [30, 82],
        "longitude": [-30, 90],
    }

    EBAS_FILTER = {
        **BASE_FILTER,
        "data_level": [None, 2],
        "set_flags_nan": True,
    }

    OBS_GROUNDBASED = {
        ##################
        #    EBAS
        ##################
        "EBAS-d-10": dict(
            obs_id="EBASratd10",
            web_interface_name="EBAS-d",
            obs_vars=["ratpm10pm25"],
            obs_vert_type="Surface",
            colocate_time=True,
            min_num_obs=DEFAULT_RESAMPLE_CONSTRAINTS,
            ts_type="daily",
            obs_filters=EBAS_FILTER,
            obs_type="ungridded",
            obs_merge_how={
                "ratpm10pm25": "eval",
            },
            obs_aux_requires={
                "ratpm10pm25": {
                    "EBASMC": [
                        "concpm10",
                        "concpm25",
                    ],
                }
            },
            obs_aux_funs={
                "ratpm10pm25":
                # variables used in computation method need to be based on AeroCom
                # units, since the colocated StationData objects (from which the
                # new UngriddedData is computed, will perform AeroCom unit check
                # and conversion)
                "(EBASMC;concpm10/EBASMC;concpm25)"
            },
            obs_aux_units={"ratpm10pm25": "1"},
        ),
        "EBAS-d-25": dict(
            obs_id="EBASratd25",
            web_interface_name="EBAS-d",
            obs_vars=["ratpm25pm10"],
            obs_vert_type="Surface",
            colocate_time=True,
            min_num_obs=DEFAULT_RESAMPLE_CONSTRAINTS,
            ts_type="daily",
            obs_filters=EBAS_FILTER,
            obs_type="ungridded",
            obs_merge_how={
                "ratpm25pm10": "eval",
            },
            obs_aux_requires={
                "ratpm25pm10": {
                    "EBASMC": [
                        "concpm10",
                        "concpm25",
                    ],
                }
            },
            obs_aux_funs={
                "ratpm25pm10":
                # variables used in computation method need to be based on AeroCom
                # units, since the colocated StationData objects (from which the
                # new UngriddedData is computed, will perform AeroCom unit check
                # and conversion)
                "(EBASMC;concpm25/EBASMC;concpm10)"
            },
            obs_aux_units={"ratpm25pm10": "1"},
        ),
        "EBAS-d-tc": dict(
            obs_id="EBASMC",
            web_interface_name="EBAS-d",
            obs_vars=[
                "concpm10",
                "concpm25",
            ],
            obs_vert_type="Surface",
            colocate_time=True,
            min_num_obs=DEFAULT_RESAMPLE_CONSTRAINTS,
            ts_type="daily",
            obs_filters=EBAS_FILTER,
        ),
    }

    # Setup for supported satellite evaluations
    OBS_SAT = {}

    OBS_CFG = {**OBS_GROUNDBASED, **OBS_SAT}

    CFG["obs_cfg"] = OBS_CFG

    return copy.deepcopy(CFG)



import xarray
from pytest import approx

from pyaerocom.projection_information import ProjectionInformation
from tests.fixtures.data_access import TEST_DATA

ROOT = TEST_DATA["MODELS"].path


def test_projection_information():
    with xarray.open_dataset(str(ROOT / "emep4no20240630" / "Base_hour.nc")) as ds:
        pi = ProjectionInformation.from_xarray(ds, "SURF_ug_PM10_rh50")
        assert pi is not None
        assert pi.x_axis == "i"
        assert pi.y_axis == "j"
        x0, y0 = float(ds["i"][0]), float(ds["j"][0])
        lon0, lat0 = float(ds["lon"][0, 0]), float(ds["lat"][0, 0])
        (lat, lon) = pi.to_latlon(x0, y0)
        assert lat == approx(lat0, abs=1e-3)
        assert lon == approx(lon0, abs=1e-3)

        x, y = pi.to_proj(lat, lon)
        assert x == approx(x0, abs=1.0)
        assert y == approx(y0, abs=1.0)



import pandas as pd
from cf_units import Unit

from pyaerocom.exceptions import UnitConversionError
from pyaerocom.time_config import SI_TO_TS_TYPE
from pyaerocom.tstype import TsType
from pyaerocom.variable_helpers import get_variable

#: default frequency for rates variables (e.g. deposition, precip)
RATES_FREQ_DEFAULT = "d"

# 1. DEFINITION OF ATOM and MOLECULAR MASSES

# Atoms
M_O = 15.999  # u
M_S = 32.065  # u
M_N = 14.0067  # u
M_H = 1.00784  # u

# Molecules
M_SO2 = M_S + 2 * M_O
M_SO4 = M_S + 4 * M_O

M_NO2 = M_N + 2 * M_O
M_NO3 = M_N + 3 * M_O

M_NH3 = M_N + 3 * M_H
M_NH4 = M_N + 4 * M_H

# Unit conversion and custom units definitions

# 2.1 Other conversion factors
HA_TO_SQM = 10000  # hectar to square metre.

# 3. LOOKUP TABLE FOR CONVERSION FACTORS

#: Custom unit conversion factors for certain variables
#: columns: variable -> from unit -> to_unit -> conversion
#: factor
UCONV_MUL_FACS = pd.DataFrame(
    [
        # ["dryso4", "mg/m2/d", "mgS m-2 d-1", M_S / M_SO4],
        # ["drynh4", "mg/m2/d", "mgN m-2 d-1", M_N/ M_NH4],
        # ["concso4", "ug S/m3", "ug m-3", M_SO4 / M_S],
        # ["SO4ugSm3", "ug/m3", "ug S m-3", M_S / M_SO4],
        # ["concso4pm25", "ug S/m3", "ug m-3", M_SO4 / M_S],
        # ["concso4pm10", "ug S/m3", "ug m-3", M_SO4 / M_S],
        ["concso2", "ug S/m3", "ug m-3", M_SO2 / M_S],
        ["concbc", "ug C/m3", "ug m-3", 1.0],
        ["concoa", "ug C/m3", "ug m-3", 1.0],
        ["concoc", "ug C/m3", "ug m-3", 1.0],
        ["conctc", "ug C/m3", "ug m-3", 1.0],
        # a little hacky for ratpm10pm25...
        # ["ratpm10pm25", "ug m-3", "1", 1.0],
        ["concpm25", "ug m-3", "1", 1.0],
        ["concpm10", "ug m-3", "1", 1.0],
        ["concno2", "ug N/m3", "ug m-3", M_NO2 / M_N],
        # ["concno3", "ug N/m3", "ug m-3", M_NO3 / M_N],
        ["concnh3", "ug N/m3", "ug m-3", M_NH3 / M_N],
        # ["concnh4", "ug N/m3", "ug m-3", M_NH4 / M_N],
        ["wetso4", "kg S/ha", "kg m-2", M_SO4 / M_S / HA_TO_SQM],
        ["concso4pr", "mg S/L", "g m-3", M_SO4 / M_S],
    ],
    columns=["var_name", "from", "to", "fac"],
).set_index(["var_name", "from"])

# may be used to specify alternative names for custom units  defined
# in UCONV_MUL_FACS

UALIASES = {
    # mass concentrations
    "ug S m-3": "ug S/m3",
    "ug C m-3": "ug C/m3",
    "ug N m-3": "ug N/m3",
    "ugC/m3": "ug C m-3",
    # deposition rates (implicit)
    ## sulphur species
    "mgS/m2": "mg S m-2",
    "mgSm-2": "mg S m-2",
    ## nitrogen species
    "mgN/m2": "mg N m-2",
    "mgNm-2": "mg N m-2",
    # deposition rates (explicit)
    ## sulphur species
    "mgS/m2/h": "mg S m-2 h-1",
    "mg/m2/h": "mg m-2 h-1",
    "mgS/m**2/h": "mg S m-2 h-1",
    "mgSm-2h-1": "mg S m-2 h-1",
    "mgSm**-2h-1": "mg S m-2 h-1",
    "mgS/m2/d": "mg S m-2 d-1",
    ## nitrogen species
    "mgN/m2/h": "mg N m-2 h-1",
    "mgN/m**2/h": "mg N m-2 h-1",
    "mgNm-2h-1": "mg N m-2 h-1",
    "mgNm**-2h-1": "mg N m-2 h-1",
    "mgN/m2/d": "mg N m-2 d-1",
    ## others
    "MM/H": "mm h-1",
    # others
    "/m": "m-1",
}


def _check_unit_endswith_freq(unit):
    """
    Check if input unit ends with an SI frequency string

    Considered SI base periods are defined in :attr:`SI_TO_TS_TYPE` (keys)
    and accepted specifications as frequency in unit string are either via
    "/<period>" or "<period>-1" (e.g. "/d" "d-1").

    Parameters
    ----------
    unit : str
        unit to be checked

    Returns
    -------
    bool
        True if input unit ends with valid frequency string, else False
    """
    if isinstance(unit, Unit):
        unit = str(unit)
    for si_unit in SI_TO_TS_TYPE:
        if unit.endswith(f"/{si_unit}") or unit.endswith(f"{si_unit}-1"):
            return True
    return False


def rate_unit_implicit(unit):
    """
    Check whether input rate unit is implicit

    Implicit rate units do not contain frequency string, e.g. "mg m-2"
    instead of "mg m-2 d-1". Such units are, e.g. used in EMEP output where
    the frequency corresponds to the output frequency, e.g. "mg m-2" per day if
    output is daily.

    Note
    ----
    For now, this is just a wrapper for :func:`_check_unit_endswith_freq`,
    but there may be more sophisticated options in the future, which may be
    added to this function.

    Parameters
    ----------
    unit : str
        unit to be tested

    Returns
    -------
    bool
        True if input unit appears to be implicit, else False.

    """
    return not _check_unit_endswith_freq(unit)


def _unit_conversion_fac_custom(var_name, from_unit):
    """Get custom conversion factor for a certain unit

    Tries to determine custom conversion factor for a variable, relative to
    that variables pyaerocom default unit. These are typically conversions
    that cannot be handled by :mod:`cf_units` (e.g. if variable is `concno3`
    which should be in units of "ug m-3" but is given in units of "ug N
    m-3", that is, nitrogen mass and not molecular NO3 mass. Since such atomar
    units are not supported by `cf_units` which is based on SI (it would think
    the "N" is Newton), pyaerocom provides a simple interface to circumvent
    these issues for such variables by providing explicit conversion factors to
    convert from e.g. "ug N m-3" to "ug m-3", for affected variables, such as
    concno3.

    Parameters
    ----------
    var_name : str
        name of variable for which factor is to be determined (needs to be
        registered in global attr. :attr:`UCONV_MUL_FACS`
    from_unit : str
        input unit (e.g. "ug N m-3")

    Raises
    ------
    UnitConversionError
        if no or no unique unit conversion factor could be retrieved for
        input from global attr. :attr:`UCONV_MUL_FACS`

    Returns
    -------
    str
        output unit
    float
        corresponding converison factor
    """
    if from_unit in UALIASES:
        from_unit = UALIASES[from_unit]
    try:
        info = UCONV_MUL_FACS.loc[(var_name, str(from_unit)), :]
        if not isinstance(info, pd.Series):
            raise UnitConversionError(
                "FATAL: Could not find unique conversion factor in table  "
                "UCONV_MUL_FACS in units_helpers.py. Please check for "
                "dulplicate entries"
            )
    except KeyError:
        raise UnitConversionError(
            f"Failed to convert unit {from_unit} (variable {var_name}). "
            f"Reason: no custom conversion factor could be inferred from table "
            f"pyaerocom.units_helpers.UCONV_MUL_FACS"
        )
    return (info.to, info.fac)


def _unit_conversion_fac_si(from_unit, to_unit):
    """Retrieve multiplication factor for unit conversion

    Works only for standard units that are supported by :mod:`cf_units`
    library. See also :func:`get_unit_conversion_factor` for more general
    cases.

    Parameters
    ----------
    from_unit : Unit or str
        input unit
    to_unit : Unit or str
        output unit

    Raises
    ------
    UnitConversionError
        if units cannot be converted into each other using cf_units package

    Returns
    --------
    float
        multiplicative conversion factor


    """
    if isinstance(from_unit, str):
        from_unit = Unit(from_unit)
    try:
        return from_unit.convert(1, to_unit)
    except ValueError:
        raise UnitConversionError(f"Failed to convert unit from {from_unit} to {to_unit}")


def _get_unit_conversion_fac_helper(from_unit, to_unit, var_name=None):
    """
    Helper for unit conversion

    Parameters
    ----------
    from_unit : str
        input unit
    to_unit : str
        output unit
    var_name : str, optional
        associated variable

    Returns
    -------
    float
        multiplication factor to convert data with input unit to output unit
        (e.g. 1000 if input unit is kg and output unit g).

    """
    pre_conv_fac = 1.0
    if from_unit == to_unit:
        # nothing to do
        return 1.0
    elif var_name is not None and var_name in UCONV_MUL_FACS.index:
        try:
            from_unit, pre_conv_fac = _unit_conversion_fac_custom(var_name, from_unit)
        except UnitConversionError:
            # from_unit is likely not custom but standard... and if not
            # call of unit_conversion_fac_si below will crash
            pass

    return _unit_conversion_fac_si(from_unit, to_unit) * pre_conv_fac


def get_unit_conversion_fac(from_unit, to_unit, var_name=None, ts_type=None):
    try:
        return _get_unit_conversion_fac_helper(from_unit, to_unit, var_name)
    except UnitConversionError:
        if (
            ts_type is not None
            and var_name is not None
            and get_variable(var_name).is_rate
            and rate_unit_implicit(from_unit)
        ):
            freq_si = TsType(ts_type).to_si()
            from_unit = f"{from_unit} {freq_si}-1"
            return _get_unit_conversion_fac_helper(from_unit, to_unit, var_name)

    raise UnitConversionError(f"failed to convert unit from {from_unit} to {to_unit}")


def convert_unit(data, from_unit, to_unit, var_name=None, ts_type=None):
    """Convert unit of data

    Parameters
    ----------
    data : np.ndarray or similar
        input data
    from_unit : cf_units.Unit or str
        current unit of input data
    to_unit : cf_units.Unit or str
        new unit of input data
    var_name : str, optional
        name of variable. If provided, and standard conversion with
        :mod:`cf_units` fails, then custom unit conversion is attempted.
    ts_type : str, optional
        frequency of data. May be needed for conversion of rate variables
        such as precip, deposition, etc, that may be defined implictly
        without proper frequency specification in the unit string.

    Returns
    -------
    data
        data in new unit
    """
    conv_fac = get_unit_conversion_fac(from_unit, to_unit, var_name, ts_type)
    if conv_fac != 1:
        data *= conv_fac
    return data



from __future__ import annotations

import iris
import numpy as np
import pytest

from pyaerocom import GriddedData
from pyaerocom._lowlevel_helpers import LayerLimits
from pyaerocom.colocation.colocation_3d import (
    ColocatedDataLists,
    colocate_vertical_profile_gridded,
)
from pyaerocom.io.read_earlinet import ReadEarlinet
from tests.fixtures.data_access import TEST_DATA

ROOT = TEST_DATA["Earlinet-test"].path

TEST_FILE: list[str] = [
    f"{ROOT}/EARLINET_AerRemSen_waw_Lev02_b0532_202109221030_202109221130_v01_qc03.nc",
]


@pytest.fixture
def fake_model_data_with_altitude():
    longitude = iris.coords.DimCoord(
        np.linspace(-15, 25, 20), var_name="lon", standard_name="longitude", units="degrees"
    )
    latitude = iris.coords.DimCoord(
        np.linspace(50, 55, 10), var_name="lat", standard_name="latitude", units="degrees"
    )
    altitude = iris.coords.DimCoord(
        np.linspace(0, 60000, 10000), var_name="alt", standard_name="altitude", units="meters"
    )
    time = iris.coords.DimCoord(
        np.arange(18892, 18892 + 7, 1),
        var_name="time",
        standard_name="time",
        units="days since epoch",
    )
    dummy = iris.cube.Cube(
        np.ones((time.shape[0], longitude.shape[0], latitude.shape[0], altitude.shape[0]))
    )

    latitude.guess_bounds()
    longitude.guess_bounds()
    altitude.guess_bounds()

    dummy.add_dim_coord(time, 0)
    dummy.add_dim_coord(longitude, 1)
    dummy.add_dim_coord(latitude, 2)
    dummy.add_dim_coord(altitude, 3)

    dummy.var_name = "bsc532aer"

    data = GriddedData(dummy)

    return data


@pytest.fixture
def example_earlinet_ungriddeddata():
    reader = ReadEarlinet()
    return reader.read(vars_to_retrieve=["bsc532aer"], files=TEST_FILE[0])


@pytest.mark.parametrize(
    "ts_type,resample_how,min_num_obs,use_climatology_ref,colocation_layer_limits,profile_layer_limits",
    [
        pytest.param(
            "daily",
            "mean",
            {"monthly": {"daily": 25}},
            False,
            ({"start": 0, "end": 6000},),
            ({"start": 0, "end": 6000},),
            id="fake_data",
        )
    ],
)
def test_colocate_vertical_profile_gridded(
    fake_model_data_with_altitude,
    example_earlinet_ungriddeddata,
    ts_type,
    resample_how,
    min_num_obs,
    use_climatology_ref,
    colocation_layer_limits: tuple[LayerLimits, ...],
    profile_layer_limits: tuple[LayerLimits, ...],
):
    colocated_data_list = colocate_vertical_profile_gridded(
        data=fake_model_data_with_altitude,
        data_ref=example_earlinet_ungriddeddata,
        ts_type=ts_type,
        resample_how=resample_how,
        min_num_obs=min_num_obs,
        use_climatology_ref=use_climatology_ref,
        colocation_layer_limits=colocation_layer_limits,
        profile_layer_limits=profile_layer_limits,
    )

    assert colocated_data_list
    assert isinstance(colocated_data_list, ColocatedDataLists)
    assert len(colocated_data_list) == 2  # col objs for statistics and viz
    assert len(colocated_data_list[0]) == len(colocation_layer_limits)
    assert len(colocated_data_list[1]) == len(profile_layer_limits)
    assert all("just_for_viz" in obj.metadata for obj in colocated_data_list[1])



from __future__ import annotations

import numpy as np
import pytest

from pyaerocom.aeroval.fairmode_stats import fairmode_stats

FAIRMODE_KEYS = {"RMSU", "sign", "crms", "bias", "rms", "alpha", "UrRV", "RV", "beta_mqi"}


@pytest.mark.parametrize(
    "obs_var,stats",
    [
        pytest.param(
            "concno2",
            dict(refdata_mean=0, refdata_std=1, data_std=1, R=1, mb=0, rms=0),
            id="dummy vars",
        )
    ],
)
def test_fairmode_stats(obs_var: str, stats: dict):
    fairmode = fairmode_stats(obs_var, stats)
    assert set(fairmode) == FAIRMODE_KEYS


@pytest.mark.parametrize(
    "obs_var,stats",
    [
        pytest.param(
            "not_a_species",
            dict(refdata_mean=0, refdata_std=1, data_std=1, R=1, mb=0, rms=0),
            id="unknown obs variable",
        ),
        pytest.param(
            "vmro3",
            dict(refdata_mean=np.nan, refdata_std=1, data_std=1, R=1, mb=0, rms=0),
            id="NaN mean",
        ),
        pytest.param(
            "conco3",
            dict(refdata_mean=0, refdata_std=np.nan, data_std=1, R=1, mb=0, rms=0),
            id="NaN obs_std",
        ),
        pytest.param(
            "conco3",
            dict(refdata_mean=0, refdata_std=1, data_std=np.nan, R=1, mb=0, rms=0),
            id="NaN mod_std",
        ),
        pytest.param(
            "conco3",
            dict(refdata_mean=0, refdata_std=1, data_std=1, R=np.nan, mb=0, rms=0),
            id="NaN R",
        ),
        pytest.param(
            "conco3",
            dict(refdata_mean=0, refdata_std=1, data_std=1, R=1, mb=np.nan, rms=0),
            id="NaN bias",
        ),
        pytest.param(
            "vmro3",
            dict(refdata_mean=0, refdata_std=1, data_std=1, R=1, mb=1, rms=np.nan),
            id="NaN rms",
        ),
    ],
)
def test_empty_stats(obs_var: str, stats: dict):
    fairmode = fairmode_stats(obs_var, stats)
    assert not fairmode


@pytest.mark.parametrize(
    "obs_var,stats,error",
    [
        pytest.param(
            "concpm10",
            {
                "refdata_mean": 0,
                "refdata_std": -1,  # Negative std
                "data_std": 1,
                "R": 1,
                "mb": 1,
                "rms": 1,
            },
            "negative obs_std=-1",
            id="obs_std",
        ),
        pytest.param(
            "concpm10",
            {
                "refdata_mean": 0,
                "refdata_std": 1,
                "data_std": -1,  # Negative std
                "R": 1,
                "mb": 1,
                "rms": 1,
            },
            "negative mod_std=-1",
            id="mod_std",
        ),
        pytest.param(
            "concpm25",
            {
                "refdata_mean": 0,
                "refdata_std": 1,
                "data_std": 1,
                "R": 10,  # Correlation must be in [-1, 1]
                "mb": 1,
                "rms": 1,
            },
            "out of range R=10",
            id="R",
        ),
    ],
)
def test_fairmode_error(obs_var: str, stats: dict, error: str):
    with pytest.raises(AssertionError) as e:
        fairmode_stats(obs_var, stats)
    assert str(e.value) == error



import json
from importlib import metadata
from pathlib import Path

import pytest
from typer.testing import CliRunner

from pyaerocom.scripts.cli import main

runner = CliRunner()


@pytest.mark.parametrize("command", ("--version", "-V"))
def test_version(command: str):
    result = runner.invoke(main, command.split())
    assert result.exit_code == 0
    assert "pyaerocom" in result.stdout
    assert metadata.version("pyaerocom") in result.stdout


@pytest.fixture()
def fake_cache(monkeypatch, tmp_path: Path):
    monkeypatch.setattr(
        "pyaerocom.io.cachehandler_ungridded.CacheHandlerUngridded.cache_dir", tmp_path
    )
    cache_file = tmp_path / "tmp.pkl"
    cache_file.write_bytes(b"")
    assert cache_file.exists()
    return tmp_path


def test_listcache(fake_cache: Path):
    assert list(fake_cache.glob("*.pkl"))
    result = runner.invoke(main, ["listcache"])
    assert result.exit_code == 0


def test_clearcache(fake_cache: Path):
    assert list(fake_cache.glob("*.pkl"))

    result = runner.invoke(main, ["clearcache"], input="y")
    # Check the clearcahce exited correctly
    assert result.exit_code == 0
    # Check that clearcache actually cleared the cache
    assert not list(fake_cache.glob("*.pkl"))


def test_clearcache_nope(fake_cache: Path):
    assert list(fake_cache.glob("*.pkl"))

    result = runner.invoke(main, ["clearcache"], input="n")
    # Check the clearcahce exited correctly
    assert result.exit_code == 0
    # Check that clearcache still there
    assert list(fake_cache.glob("*.pkl"))


def test_browse():
    result = runner.invoke(
        main, ["browse", "EARLINET"]
    )  # EARLINET is an arbitrary choice, but it works. Could think of better choice.
    print(result.stdout)
    assert result.exit_code == 0


def test_ppiaccess():
    result = runner.invoke(main, ["ppiaccess"])
    assert result.exit_code == 0


def test_init():
    result = runner.invoke(main, ["init"])
    assert result.exit_code == 0


@pytest.fixture()
def config_json(monkeypatch, tmp_path: Path, eval_config: dict) -> Path:
    def do_not_run(self, model_name=None, obs_name=None, var_list=None, update_interface=True):
        assert model_name is None
        assert obs_name is None
        assert var_list is None
        assert update_interface is True

    monkeypatch.setattr("pyaerocom.scripts.cli.ExperimentProcessor.run", do_not_run)

    path = tmp_path / "conf.json"
    path.write_text(json.dumps(eval_config))
    return path


@pytest.mark.parametrize("cfg", ("cfgexp1",))
def test_aeroval(config_json: Path):
    assert config_json.is_file()
    result = runner.invoke(main, ["aeroval", str(config_json)])
    assert result.exit_code == 0



from __future__ import annotations

from itertools import chain
from pathlib import Path

import pytest
import simplejson

from pyaerocom.io.ebas_file_index import EbasFileIndex
from pyaerocom.io.ebas_nasa_ames import EbasNasaAmesFile
from pyaerocom.io.read_ebas import ReadEbas

from .data_access import DataForTests

EBAS_FILEDIR = DataForTests("obsdata/EBASMultiColumn/data").path
ebas_info_file = DataForTests("scripts/ebas_files.json").path
assert ebas_info_file.exists()
EBAS_FILES = simplejson.loads(ebas_info_file.read_text())
for sites in EBAS_FILES.values():
    for files in sites.values():
        for file in files:
            assert (EBAS_FILEDIR / file).exists()

EBAS_ISSUE_FILES = {
    # conco3 tower - 3 different measurement heights
    "o3_tower": "CZ0003R.20150101000000.20181107114213.uv_abs.ozone.air.1y.1h..CZ06L_uv_abs.lev2.nas",
    # conco3 - Neg. meas periods
    "o3_neg_dt": "NZ0003G.20090110030000.20181130115605.uv_abs.ozone.air.9h.1h.US06L_Thermo_49C_LAU.US06L_AM.lev2.nas",
    # conco3 - Most common meas period is 150s and does not correspond to any of the supported base frequencies ['minutely', 'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'native']
    "o3_tstype": "LT0015R.20080101000000.20081231000000.uv_abs.ozone.air.15d.1h.LT01L_uv_abs_15.LT01L_uv_abs..nas",
    # concpm10 - could not resolve unique data column for concpm10 (EBAS varname: ['pm10_mass'])
    "pm10_colsel": "ID1013R.20180101000000.20200102000000.beta_gauge_particulate_sampler.pm10_mass.pm10.1y.1h.ID01L_MetOne_BAM1020..lev2.nas",
    # concpm10 Aliartos - Most common meas period is 172800s and does not correspond to any of the supported base frequencies ['minutely', 'hourly', 'daily', 'weekly', 'monthly', 'yearly', 'native']
    "pm10_tstype": "GR0001R.20060119000000.20080120000000.b-attenuation.pm10_mass.pm10.11mo.1w.GR01L_b_att_01.GR01L_b-attenuation..nas",
    # sc550dryaer Jungfraujoch - only first file
    "Jungfraujoch": EBAS_FILES["sc550dryaer"]["Jungfraujoch"][0],
}


@pytest.fixture
def ebas_issue_files(issue_files: str) -> Path:
    """file path for `EBAS_ISSUE_FILES[issue_files]`"""
    return EBAS_FILEDIR / EBAS_ISSUE_FILES[issue_files]


def get_ebas_filelist(var_name: str) -> list[Path]:
    """all file paths for `var_name`"""
    paths: list[Path] = [
        EBAS_FILEDIR / file for files in EBAS_FILES[var_name].values() for file in files
    ]
    assert all(path.exists() for path in paths)
    return paths


@pytest.fixture
def ebas_files(file_vars: list[str] | str | None) -> list[Path] | None:
    if file_vars is None:
        return None
    if isinstance(file_vars, str):
        return get_ebas_filelist(file_vars)

    paths = (get_ebas_filelist(var_name) for var_name in file_vars)
    return list(chain.from_iterable(paths))


@pytest.fixture(scope="module")
def ebas() -> EbasFileIndex:
    """ebas file index for test DB"""
    EBAS_SQLite_DB = EBAS_FILEDIR.parent / "ebas_file_index.sqlite3"
    assert EBAS_SQLite_DB.exists()
    return EbasFileIndex(EBAS_SQLite_DB)


@pytest.fixture(scope="session")
def data_scat_jungfraujoch():
    """subset of sc550aer obs from Jungfraujoch"""
    reader = ReadEbas("EBASSubset")
    return reader.read("sc550aer", station_names="Jungfrau*")


@pytest.fixture(scope="session")
def data_scat_jungfraujoch_full():
    """all sc550aer obs from Jungfraujoch"""
    reader = ReadEbas()
    return reader.read("sc550aer", station_names="Jungfrau*")


@pytest.fixture(scope="session")
def loaded_nasa_ames_example() -> EbasNasaAmesFile:
    """loaded sc550dryaer Jungfraujoch file"""
    path = EBAS_FILEDIR / EBAS_ISSUE_FILES["Jungfraujoch"]
    assert path.exists()
    return EbasNasaAmesFile(path)



from pyaerocom.stats.types import StatsDict


class FilterDropStats:
    """
    Drops stats from a StatsDict which are in a provided list of values.
    """

    def __init__(self, stats_to_drop: list[str]):
        self.stats_to_drop = stats_to_drop

    def __call__(self, stats: StatsDict) -> StatsDict:
        for k in self.stats_to_drop:
            if k in stats.keys():
                del stats[k]

        return stats



from __future__ import annotations

import logging
import multiprocessing as mp
from copy import deepcopy
from datetime import date, datetime
from pathlib import Path
from typing import Optional

import typer

from pyaerocom import const
from pyaerocom.scripts.cams2_83.config import CFG
from pyaerocom.scripts.cams2_83.evaluation import EvalType, runnermedianscores, runnermos

app = typer.Typer(add_completion=False, no_args_is_help=True)
logger = logging.getLogger(__name__)


def make_config_mos(
    start_date: date,
    end_date: date,
    data_path: Path,
    coldata_path: Path,
    eval_type: EvalType,
    id: str,
    name: str,
    description: str,
) -> dict:
    logger.info("Making the configuration")

    models = ["ENS", "MOS"]

    cfg = deepcopy(CFG)
    cfg.update(
        model_cfg={
            f"{model}": dict(
                model_id=f"CAMS2-83.{model}.day0.FC",
                model_kwargs=dict(
                    daterange=[f"{start_date}", f"{end_date}"],
                ),
            )
            for model in models
        },
        periods=eval_type.periods(start_date, end_date),
        json_basedir=str(data_path),
        coldata_basedir=str(coldata_path),
    )

    if eval_type is not None:
        eval_type.check_dates(start_date, end_date)
        cfg.update(eval_type.freqs_config())

    cfg.update(only_json=True)

    cfg.update(exp_id=id, exp_name=name, exp_descr=description)

    cfg.update(use_fairmode=True)

    return cfg


@app.command()
def main(
    eval_type: EvalType = typer.Argument(...),
    start_date: datetime = typer.Argument(
        ..., formats=["%Y-%m-%d", "%Y%m%d"], help="evaluation start date"
    ),
    end_date: datetime = typer.Argument(
        ..., formats=["%Y-%m-%d", "%Y%m%d"], help="evaluation end date"
    ),
    data_path: Path = typer.Option(
        Path("../../data").resolve(),
        exists=True,
        readable=True,
        writable=True,
        help="where results are stored",
    ),
    coldata_path: Path = typer.Option(
        Path("../../coldata").resolve(),
        exists=True,
        readable=True,
        writable=True,
        help="where colocated data are stored, this is useless here but this path needs to exist",
    ),
    cache: Optional[Path] = typer.Option(
        None,
        help="Optional path to cache. If nothing is given, the default pyaerocom cache is used",
    ),
    id: str = typer.Option(CFG["exp_id"], help="experiment ID"),
    name: str = typer.Option(CFG["exp_name"], help="experiment name"),
    description: str = typer.Option(CFG["exp_descr"], help="experiment description"),
    pool: int = typer.Option(
        1,
        "--pool",
        "-p",
        min=1,
        help="CPUs for running the median scores",
    ),
):
    if pool > mp.cpu_count():
        logger.warning(
            f"The given pool {pool} is larger than the maximum CPU count {mp.cpu_count()}. Using that instead"
        )
        pool = 1

    cfg = make_config_mos(
        start_date,
        end_date,
        data_path,
        coldata_path,
        eval_type,
        id,
        name,
        description,
    )

    # we do not want the cache produced in previous runs to be silently cleared
    const.RM_CACHE_OUTDATED = False

    logger.info("Standard run")
    runnermos(cfg, cache, dry_run=False)
    logger.info("Special run for median scores only")
    runnermedianscores(cfg, cache, dry_run=False, analysis=False, pool=pool)


if __name__ == "__main__":
    main()



from __future__ import annotations

import cartopy.mpl.geoaxes
import numpy as np
import pytest
from matplotlib.figure import Figure

from pyaerocom import GriddedData
from pyaerocom.colocation.colocated_data import ColocatedData
from pyaerocom.config import ALL_REGION_NAME
from pyaerocom.exceptions import DataDimensionError
from pyaerocom.plot.config import ColorTheme, get_color_theme
from pyaerocom.plot.mapping import (
    get_cmap_maps_aerocom,
    init_map,
    plot_griddeddata_on_map,
    plot_map_aerocom,
    plot_nmb_map_colocateddata,
    set_map_ticks,
)


@pytest.mark.parametrize(
    "color_theme,vmin,vmax,name",
    [
        (None, None, None, "Blues"),
        (None, -1, 1, "shiftedcmap"),
        (get_color_theme("light"), None, None, "Blues"),
        (get_color_theme("dark"), None, None, "viridis"),
    ],
)
def test_get_cmap_maps_aerocom(color_theme, vmin, vmax, name):
    cmap = get_cmap_maps_aerocom(color_theme, vmin, vmax)
    assert cmap.name == name


def test_set_cmap_ticks():
    ax = init_map()
    ax = set_map_ticks(ax, None, None)
    assert isinstance(ax, cartopy.mpl.geoaxes.GeoAxes)
    xticks = [-90, 0, 90]
    yticks = [-60, 5, 60]
    ax = set_map_ticks(ax, xticks=xticks, yticks=yticks)
    assert list(ax.get_xticks()) == xticks
    assert list(ax.get_yticks()) == yticks


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(),
        dict(projection="PlateCarree"),
        dict(projection="Mercator"),
        dict(projection="Miller"),
        dict(fig=Figure()),
        dict(fix_aspect=True),
        dict(color_theme="dark"),
        dict(color_theme=ColorTheme("dark")),
    ],
)
def test_init_map(kwargs: dict):
    ax = init_map(**kwargs)
    assert isinstance(ax, cartopy.mpl.geoaxes.GeoAxes)


@pytest.mark.parametrize(
    "kwargs,error",
    [
        pytest.param(
            dict(projection="blaaa"),
            "no such projection blaaa",
            id="unknown projection",
        ),
        pytest.param(
            dict(projection=42),
            "Input for projection needs to be instance of cartopy.crs.Projection",
            id="wrong projection",
        ),
    ],
)
def test_init_map_error(kwargs: dict, error: str):
    with pytest.raises(ValueError) as e:
        init_map(**kwargs)
    assert str(e.value) == error


@pytest.fixture
def gridded_data(data_tm5: GriddedData, key: GriddedData | str | None) -> GriddedData:
    if key is None:
        return data_tm5

    data: GriddedData = data_tm5.copy()
    if key == "allnan":
        data.grid.data = data.grid.data * np.nan
        return data

    if key == "negval":
        data.grid.data[0, 0, 0] = -100
        return data

    if key == "const":
        data.grid.data[:, :, :] = 1
        return data

    return key  # type:ignore[return-value]


@pytest.mark.parametrize(
    "key,kwargs",
    [
        (None, dict()),
        (None, dict(add_cbar=False)),
        (None, dict(cbar_levels=[0.2, 0.4, 0.6])),
        (None, dict(cbar_levels=[0.2, 0.6], cmap="Blues")),
        (None, dict(cbar_levels=[0.2, 0.6], add_zero=True)),
        (None, dict(vmin=-0.2, log_scale=True)),
        (None, dict(vmin=-0.2, log_scale=False)),
        (None, dict(vmin=-0.1, log_scale=True, vmax=2, cmap="Reds")),
        (None, dict(discrete_norm=False)),
        (None, dict(vmin=-0.1, discrete_norm=False)),
        ("negval", dict(discrete_norm=False)),
        (None, dict(log_scale=False, add_zero=True)),
        (None, dict(log_scale=False, cmap="viridis")),
        (None, dict(c_over="r", c_under="b")),
        (None, dict(cbar_levels=[0.2, 0.6], c_over="r", c_under="b")),
        (None, dict(cbar_levels=[0.2, 0.6], c_over="r")),
        (None, dict(add_cbar=True, var_name="od550aer", unit="ug")),
        (None, dict(add_cbar=True, cbar_ticks=[0.1, 0.2, 0.3])),
        (None, dict(add_cbar=True, cbar_ticks=[0.1, 0.2, 0.3], cbar_ticks_sci=True)),
    ],
)
@pytest.mark.filterwarnings("ignore:More than 20 figures have been opened:RuntimeWarning")
def test_plot_griddeddata_on_map(gridded_data: GriddedData, kwargs: dict):
    val = plot_griddeddata_on_map(gridded_data, **kwargs)
    assert isinstance(val, Figure)


class Fake4D(GriddedData):
    has_latlon_dims = True
    ndim = 4


@pytest.mark.parametrize(
    "key,kwargs,exception,error",
    [
        pytest.param(
            42,
            dict(),
            ValueError,
            "need GriddedData",
            id="no GriddedData",
        ),
        pytest.param(
            GriddedData(),
            dict(),
            DataDimensionError,
            "Input data needs to have latitude and longitude dimension",
            id="no lon/lat",
        ),
        pytest.param(
            Fake4D(),
            dict(),
            DataDimensionError,
            "Input data needs to be 2 dimensional or 3D with time being the 3rd dimension",
            id="fake 4D",
        ),
        pytest.param(
            None,
            dict(ax=42),
            ValueError,
            "Invalid input for ax, need GeoAxes",
            id="wrong ax",
        ),
        pytest.param(
            None,
            dict(cbar_levels=[0.2, 0.4, 0.6], vmin=0.1, vmax=0.2),
            ValueError,
            "Please provide either vmin/vmax OR cbar_levels",
            id="cbar_levels and vmin/vmax",
        ),
        pytest.param(
            "const",
            dict(discrete_norm=False),
            ValueError,
            "Minimum value in data equals maximum value: 1.0",
            id="constant",
        ),
        pytest.param(
            "allnan",
            dict(),
            ValueError,
            "Cannot plot map of data: all values are NaN",
            id="all NaN",
        ),
    ],
)
@pytest.mark.filterwarnings("ignore:More than 20 figures have been opened:RuntimeWarning")
def test_plot_griddeddata_on_map_error(
    gridded_data: GriddedData, kwargs: dict, exception: type[Exception], error: str
):
    with pytest.raises(exception) as e:
        plot_griddeddata_on_map(gridded_data, **kwargs)
    assert str(e.value) == error


@pytest.mark.parametrize(
    "region",
    [
        ALL_REGION_NAME,
        "EUROPE",
        pytest.param(
            "EEUROPE",
            marks=pytest.mark.filterwarnings("ignore:Out of bound index found:DeprecationWarning"),
        ),
    ],
)
def test_plot_map_aerocom(data_tm5: GriddedData, region: str):
    val = plot_map_aerocom(data_tm5, region)
    assert isinstance(val, Figure)


def test_plot_map_aerocom_error():
    with pytest.raises(ValueError):
        plot_map_aerocom(42, ALL_REGION_NAME)


def test_plot_nmb_map_colocateddata(coldata_tm5_aeronet: ColocatedData):
    val = plot_nmb_map_colocateddata(coldata_tm5_aeronet)
    assert isinstance(val, cartopy.mpl.geoaxes.GeoAxes)


def test_plot_nmb_map_colocateddata4D(coldata_tm5_tm5: ColocatedData):
    val = plot_nmb_map_colocateddata(coldata_tm5_tm5)
    assert isinstance(val, cartopy.mpl.geoaxes.GeoAxes)


@pytest.mark.parametrize(
    "coldataset,exception,error",
    [
        pytest.param(
            "fake_nodims",
            AssertionError,
            "",
            id="nodims",
        ),
    ],
)
def test_plot_nmb_map_colocateddataFAIL(
    coldata: ColocatedData, exception: type[Exception], error: str
):
    with pytest.raises(exception) as e:
        plot_nmb_map_colocateddata(coldata)
    assert str(e.value) == error



import warnings
import cartopy

from pyaerocom.plot.mapping import init_map


def plot_coordinates(
    lons,
    lats,
    xlim=None,
    ylim=None,
    label=None,
    legend=True,
    color=None,
    marker=None,
    markersize=8,
    ax=None,
    **kwargs,
):
    """Plot input coordinates on a map

    lons : ndarray
        array of longitude coordinates (can also be list or tuple)
    lats : ndarray
        array of latitude coordinates (can also be list or tuple)
    xlim : tuple
        longitude range
    ylim : tuple
        latitude range
    label : str, optional
        label of data
    legend : bool
        whether or not to display a legend, defaults to True.
    color : str, optional
        color of markers, defaults to red
    marker : str, optional
        marker shape, defaults to 'o'
    markersize : int
        size of markers
    ax : GeoAxes
        axes instance to be plotted into
    **kwargs
        additional keyword args passed on to :func:`init_map`

    Returns
    -------
    GeoAxes

    """
    warnings.warn(
        "matplotlib based plotting is no longer directly supported. This function may be removed in future versions.",
        DeprecationWarning,
    )

    if xlim is None:
        xlim = (-180, 180)
    if ylim is None:
        ylim = (-90, 90)
    if color is None:
        color = "r"
    if marker is None:
        marker = "o"

    if not isinstance(ax, cartopy.mpl.geoaxes.GeoAxes):
        ax = init_map(xlim, ylim, ax=ax, **kwargs)

    if label is None:
        label = f"{len(lons)} stations"

    ax.scatter(lons, lats, markersize, marker=marker, color=color, label=label)

    if legend and label:
        ax.legend()

    return ax



import pytest

from pyaerocom import const
from pyaerocom.io import ReadPyaro, ReadUngridded


def test_invalid_init_data_dirs():
    data_dirs = "/bla/blub"
    with pytest.raises(ValueError) as e:
        ReadUngridded(["EBASMC", "GHOST.EEA.daily"], data_dirs=data_dirs)
    assert str(e.value) == f"Invalid input for data_dirs ({data_dirs}); needs to be a dictionary."


def test_supported():
    supported_datasets = ReadUngridded().supported_datasets
    assert len(supported_datasets) >= 12
    datasets = {
        "AeronetInvV3Lev2.daily",
        "AeronetInvV3Lev1.5.daily",
        "AeronetInvV3L2Subset.daily",
        "AeronetSDAV3Lev1.5.daily",
        "AeronetSDAV3Lev2.daily",
        "AeronetSDAV3L2Subset.daily",
        "AeronetSunV3Lev1.5.daily",
        "AeronetSunV3Lev1.5.AP",
        "AeronetSunV3Lev2.daily",
        "AeronetSunV3Lev2.AP",
        "AeronetSunV3L2Subset.daily",
        "EARLINET",
        "EBASMC",
        "EBASSubset",
        "DMS_AMS_CVO",
        "GAWTADsubsetAasEtAl",
        "GHOST.EEA.monthly",
        "GHOST.EEA.hourly",
        "GHOST.EEA.daily",
        "GHOST.EBAS.monthly",
        "GHOST.EBAS.hourly",
        "GHOST.EBAS.daily",
    }
    assert datasets <= set(supported_datasets)


@pytest.mark.parametrize("data_ids", [None, "Blaaaaaa"])
@pytest.mark.parametrize("ignore_cache", [False, True])
def test_ReadUngridded___init__(data_ids, ignore_cache):
    _caching = const.CACHING
    ReadUngridded(data_ids=data_ids, ignore_cache=ignore_cache)
    if ignore_cache:
        assert not const.CACHING
    const.CACHING = _caching


@pytest.mark.parametrize(
    "filter_post,nst,nmeta",
    [
        (None, 22, 22),
        (dict(station_name="La_Paz"), 1, 1),
        (dict(station_name=["La_Paz", "AAO*"]), 2, 2),
        (dict(altitude=[1000, 10000]), 3, 3),
        (
            dict(altitude=[1000, 10000], ignore_station_names=dict(od550aer="La_Paz")),
            2,
            2,
        ),
        (dict(altitude=[1000, 10000], ignore_station_names="La_*"), 2, 2),
        (
            dict(altitude=[1000, 10000], ignore_station_names=["La_*", "Mauna_Loa"]),
            1,
            1,
        ),
    ],
)
@pytest.mark.parametrize(
    "only_cached,caching",
    [(False, False), (True, True), (False, True)],
)
def test_ReadUngridded_read(only_cached, filter_post, nst, nmeta, caching):
    reader = ReadUngridded()

    const.CACHING = caching
    data = reader.read(
        data_ids="AeronetSunV3L2Subset.daily",
        vars_to_retrieve="od550aer",
        only_cached=only_cached,
        filter_post=filter_post,
    )

    assert len(data.metadata) == nmeta
    assert len(data.unique_station_names) == nst

    const.CACHING = True


def test_ReadUngridded_read_error():
    reader = ReadUngridded()

    with pytest.raises(NotImplementedError) as e:
        reader.read(
            data_ids="AeronetSunV3L2Subset.daily",
            vars_to_retrieve=["od550aer", "ang4487aer"],
            filter_post=dict(altitude=[1000, 10000], ignore_station_names=dict(od550aer="La_Paz")),
        )
    assert str(e.value).startswith(
        "Cannot filter different sites for multivariable UngriddedData objects"
    )


def test_basic_attributes():
    reader = ReadUngridded()
    assert not reader.ignore_cache
    assert reader.data_ids == []
    with pytest.raises(ValueError):
        reader.get_lowlevel_reader()
    with pytest.raises(AttributeError):
        reader.dataset_provides_variables()


#########################################
# Tests for use of PyaroConfig
#########################################


def test_init_configs(pyaro_testconfig):
    reader = ReadUngridded(configs=pyaro_testconfig)

    for i, reader_id in enumerate(reader.data_ids):
        assert reader_id == pyaro_testconfig[i].name


def test_get_lowlevel_reader_configs(pyaro_testconfig):
    reader = ReadUngridded(configs=pyaro_testconfig)

    ll_reader = reader.get_lowlevel_reader(data_id=pyaro_testconfig[0].name)
    assert isinstance(ll_reader, ReadPyaro)


def test_supported_pyaro(pyaro_testconfig):
    reader = ReadUngridded(configs=pyaro_testconfig)

    assert ReadPyaro in reader.SUPPORTED_READERS
    assert pyaro_testconfig[0].reader_id in reader.supported_datasets


def test_read_pyaro_and_other(pyaro_testconfig):
    data_ids = ["AeronetInvV3L2Subset.daily"]
    reader = ReadUngridded(data_ids=data_ids)

    data = reader.read(configs=pyaro_testconfig)
    assert len(data.contains_datasets) == 3
    assert pyaro_testconfig[0].name in data.contains_datasets


def test_read_pyaro_valid_data(pyaro_testconfig):
    reader = ReadUngridded(configs=pyaro_testconfig)

    data = reader.read(vars_to_retrieve=["concso4"])
    station = data.to_station_data(0)

    assert not station["concso4"].empty
    assert not station["concso4"].hasnans

    station = data.to_station_data(1)

    assert not station["concso4"].empty
    assert not station["concso4"].hasnans


def test_config_map(pyaro_testconfig):
    reader = ReadUngridded(configs=pyaro_testconfig)
    m = reader.config_map
    assert m[pyaro_testconfig[0].name] == pyaro_testconfig[0]
    assert m[pyaro_testconfig[1].name] == pyaro_testconfig[1]


def test_get_vars_supported_pyaro(pyaro_testconfig):
    config = pyaro_testconfig[0]
    reader = ReadUngridded(configs=config)
    assert reader.get_vars_supported(obs_id=config.name, vars_desired=["concso4"])


##
# Tests that raises errors
##


def test_config_name_already_exists(pyaro_testconfig):
    data_ids = ["AeronetInvV3L2Subset.daily"]
    config = pyaro_testconfig[0]
    config.name = data_ids[0]

    with pytest.raises(NameError, match="cannot have the same name as an included dataset"):
        ReadUngridded(configs=pyaro_testconfig)

    with pytest.raises(NameError, match="cannot have the same name as an included dataset"):
        ReadUngridded(data_ids=data_ids).read(configs=config)



import numpy as np
import pytest
from matplotlib.axes import Axes

import pyaerocom.plot.plotscatter as mod


def test_plot_scatter():
    val = mod.plot_scatter([1, 2], [1, 2])
    assert isinstance(val, Axes)


X, Y = np.arange(100), np.arange(100) * 2


@pytest.mark.parametrize(
    "args",
    [
        (dict(x_vals=np.ones(10), y_vals=np.ones(10))),
        (
            dict(
                x_vals=X,
                y_vals=Y,
                var_name="od550aer",
                var_name_ref="bla",
                x_name="OBS",
                y_name="MODEL",
                start=np.datetime64("2010-01-01"),
                stop=np.datetime64("2010-12-31"),
                ts_type="monthly",
                unit="ONE",
                stations_ok=10,
                filter_name="BLAAAAA",
                lowlim_stats=10,
                highlim_stats=90,
                loglog=True,
                figsize=(30, 30),
                fontsize_base=14,
                fontsize_annot=13,
                marker="o",
                color="lime",
                alpha=0.1,
            )
        ),
        (dict(x_vals=np.arange(-10, 10), y_vals=np.arange(-5, 15), loglog=True)),
    ],
)
def test_plot_scatter_aerocom(args):
    val = mod.plot_scatter_aerocom(**args)
    assert isinstance(val, Axes)



from __future__ import annotations

import logging
import re
import time
import warnings
from pathlib import Path
from reprlib import repr

import numpy as np
import xarray as xr

from pyaerocom import ColocatedData
from pyaerocom.aeroval._processing_base import ProcessingEngine
from pyaerocom.aeroval.coldatatojson_helpers import _select_period_season_coldata, init_regions_web
from pyaerocom.exceptions import DataCoverageError, UnknownRegion
from pyaerocom.io.cams2_83.models import ModelName

logger = logging.getLogger(__name__)
warnings.filterwarnings("ignore")


class CAMS2_83_Engine(ProcessingEngine):
    def run(self, files: list[list[str | Path]], var_list: list) -> None:  # type:ignore[override]
        logger.info(f"Processing: {repr(files)}")
        coldata = [ColocatedData(data=file) for file in files]
        coldata, found_vars = self._sort_coldata(coldata)
        start = time.time()
        if var_list is None:
            var_list = list(found_vars)
        for var in var_list:
            logger.info(f"Processing Component: {var}")
            self.process_coldata(coldata[var])

        logger.info(f"Time for weird plot: {time.time() - start} sec")

    def process_coldata(self, coldata: list[ColocatedData]) -> None:
        use_weights = self.cfg.statistics_opts.weighted_stats
        out_dirs = self.cfg.path_manager.get_json_output_dirs(True)
        forecast_days = self.cfg.statistics_opts.forecast_days
        periods = self.cfg.time_cfg.periods

        if "var_name_input" in coldata[0].metadata:
            obs_var = coldata[0].metadata["var_name_input"][0]
            model_var = coldata[0].metadata["var_name_input"][1]
        else:
            obs_var = model_var = "UNDEFINED"

        # for the MOS/ENS evaluation experiment the models are just strings
        # we do not want them added to the ModelName class
        # so we need a bunch of ugly special cases here
        modelname = coldata[0].model_name.split("-")[2]
        if modelname == "ENS" or modelname == "MOS":
            model = modelname
        else:
            model = ModelName[modelname]
        vert_code = coldata[0].get_meta_item("vert_code")
        obs_name = coldata[0].obs_name
        if modelname == "ENS" or modelname == "MOS":  # MOS/ENS evaluation special case
            mcfg = self.cfg.model_cfg.get_entry(modelname)
        else:
            mcfg = self.cfg.model_cfg.get_entry(model.name)
        var_name_web = mcfg.get_varname_web(model_var, obs_var)
        seasons = self.cfg.time_cfg.get_seasons()

        regions_how = "country"
        use_country = True
        for i in range(forecast_days):
            coldata[i].data["season"] = coldata[i].data.time.dt.season
            (regborders, regs, regnames) = init_regions_web(coldata[i], regions_how)

        results = {}

        for regid, regname in regnames.items():
            results[regname] = {}
            logger.info(f"Creating subset for {regname}")
            try:
                subset_region = [
                    col.filter_region(regid, check_country_meta=use_country) for col in coldata
                ]
            except (DataCoverageError, UnknownRegion) as e:
                logger.info(f"Skipping forecast plot for {regname} due to error {str(e)}")
                continue
            for per in periods:
                for season in seasons:
                    perstr = f"{per}-{season}"

                    stats_list: dict[str, list[float]] = dict(
                        rms=[], R=[], nmb=[], mnmb=[], fge=[]
                    )
                    logger.info(f"Making subset for {regid}, {per} and {season}")
                    if season not in coldata[0].data["season"].data and season != "all":
                        logger.info(
                            f"Season {season} is not available for {per} and will be skipped"
                        )
                        continue

                    try:
                        subset = [
                            _select_period_season_coldata(col, per, season)
                            for col in subset_region
                        ]
                    except (DataCoverageError, UnknownRegion) as e:
                        logger.info(f"Skipping forecast plot due to error {str(e)}")
                        continue

                    for forecast_hour in range(24 * forecast_days):
                        logger.debug(f"Calculating statistics for hour {forecast_hour}")
                        leap, hour = divmod(forecast_hour, 24)
                        ds = subset[leap]
                        ds = ds.data.sel(time=(ds.time.dt.hour == hour))
                        start = time.time()
                        stats = self._get_median_stats_point_vec(ds, use_weights)
                        logger.debug(time.time() - start)
                        for key in stats_list:
                            stats_list[key].append(stats[key])

                    out_dirs = self.cfg.path_manager.get_json_output_dirs(True)  # noqa: F841

                    results[f"{regname}"][f"{perstr}"] = stats_list

            self.exp_output.add_forecast_entry(
                results[regname],
                regname,
                obs_name,
                var_name_web,
                vert_code,
                (
                    modelname if (modelname == "ENS" or modelname == "MOS") else model.name
                ),  # MOS/ENS evaluation special case
                model_var,
            )

    def _get_median_stats_point(self, data: xr.DataArray, use_weights: bool) -> dict[str, float]:
        stats_list: dict[str, list[float]] = dict(rms=[], R=[], nmb=[], mnmb=[], fge=[])
        station_list = data.station_name.data
        for station in station_list:
            d = data.sel(station_name=[station])
            arr = ColocatedData(data=d)
            stats = arr.calc_statistics(use_area_weights=use_weights)
            for key in stats_list.keys():
                stats_list[key].append(stats[key])
        median_stats = {}
        for key in stats_list.keys():
            median_stats[key] = np.nanmedian(np.array(stats_list[key]))

        return median_stats

    def _get_median_stats_point_vec(
        self, data: xr.DataArray, use_weights: bool
    ) -> dict[str, float]:
        stats_list: dict[str, float] = dict(rms=0.0, R=0.0, nmb=0.0, mnmb=0.0, fge=0.0)

        obsvals = data.data[0]
        modvals = data.data[1]

        diff = modvals - obsvals
        diffsquare = diff**2
        sum_obs = np.nansum(obsvals, axis=0)
        sum_diff = np.nansum(diff, axis=0)
        sum_vals = obsvals + modvals

        tmp = diff / sum_vals

        nmb = np.where(sum_obs == 0, np.nan, sum_diff / sum_obs)

        mnmb = 2.0 * np.nanmean(tmp, axis=0)
        fge = 2.0 * np.nanmean(np.abs(tmp), axis=0)
        rms = np.sqrt(np.nanmean(diffsquare, axis=0))

        R = self._pearson_R_vec(obsvals, modvals)

        stats_list["rms"] = np.nanmedian(rms)
        stats_list["R"] = np.nanmedian(R)
        stats_list["nmb"] = np.nanmedian(nmb)
        stats_list["mnmb"] = np.nanmedian(mnmb)
        stats_list["fge"] = np.nanmedian(fge)

        return stats_list

    def _pearson_R_vec(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        xmean = np.nanmean(x, axis=0)
        ymean = np.nanmean(y, axis=0)
        xm = x - xmean
        ym = y - ymean
        normxm = np.sqrt(np.nansum(xm * xm, axis=0))
        normym = np.sqrt(np.nansum(ym * ym, axis=0))

        r = np.where(
            normxm * normym == 0.0,
            np.nan,
            np.nansum(xm * ym, axis=0) / (normxm * normym),
        )

        return r

    def _sort_coldata(
        self, coldata: list[ColocatedData]
    ) -> tuple[dict[str, list[ColocatedData]], set[str]]:
        col_dict = dict()
        var_list = []

        for col in coldata:
            obs_var = col.metadata["var_name_input"][0]

            if obs_var in col_dict:
                col_dict[obs_var].append(col)
            else:
                col_dict[obs_var] = [col]
                var_list.append(obs_var)

        for var, cols in col_dict.items():
            col_dict[var] = sorted(cols, key=lambda x: self._get_day(x.model_name))

        return col_dict, set(var_list)

    def _get_day(self, model_name: str) -> int:
        return int(re.search(".*day([0-3]).*", model_name).group(1))
